from itertools import chain
from django.db import migrations, models
import django.db.models.deletion
import leprikon.models.fields


from leprikon import migrations as leprikon_migrations


def migrate_cms_apps(apps, schema_editor):
    Page = apps.get_model("cms", "Page")

    Page.objects.filter(application_urls="LeprikonSubjectTypeApp").update(application_urls="LeprikonActivityTypeApp")


class Migration(migrations.Migration):

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        ("leprikon", "0087_alter_schoolyear_year"),
    ]

    operations = (
        [
            migrations.RunPython(migrate_cms_apps, migrations.RunPython.noop),
            migrations.RenameModel(old_name="SubjectType", new_name="ActivityType"),
            migrations.AlterModelOptions(
                name="activitytype",
                options={
                    "ordering": ("order",),
                    "verbose_name": "activity type",
                    "verbose_name_plural": "activity types",
                },
            ),
            migrations.RenameField(
                model_name="activitytype",
                old_name="subject_type",
                new_name="model",
            ),
            migrations.AlterField(
                model_name="activitytype",
                name="model",
                field=models.CharField(
                    choices=[("course", "course"), ("event", "fixed time event"), ("orderable", "orderable event")],
                    max_length=10,
                    verbose_name="data model",
                ),
            ),
            migrations.CreateModel(
                name="ActivityTypeTexts",
                fields=[],
                options={
                    "verbose_name": "texts",
                    "verbose_name_plural": "texts",
                    "proxy": True,
                    "indexes": [],
                    "constraints": [],
                },
                bases=("leprikon.activitytype",),
            ),
            migrations.CreateModel(
                name="ActivityTypePrintSetups",
                fields=[],
                options={
                    "verbose_name": "print setups",
                    "verbose_name_plural": "print setups",
                    "proxy": True,
                    "indexes": [],
                    "constraints": [],
                },
                bases=("leprikon.activitytype",),
            ),
            migrations.AlterField(
                model_name="courselistplugin",
                name="course_types",
                field=models.ManyToManyField(
                    blank=True,
                    help_text="Keep empty to skip searching by course types.",
                    limit_choices_to={"activity_type": "course"},
                    related_name="+",
                    to="leprikon.ActivityType",
                    verbose_name="course types",
                ),
            ),
            migrations.AlterField(
                model_name="eventlistplugin",
                name="event_types",
                field=models.ManyToManyField(
                    blank=True,
                    help_text="Keep empty to skip searching by event types.",
                    limit_choices_to={"activity_type": "event"},
                    related_name="+",
                    to="leprikon.ActivityType",
                    verbose_name="event types",
                ),
            ),
            migrations.AlterField(
                model_name="filteredcourselistplugin",
                name="course_types",
                field=models.ManyToManyField(
                    limit_choices_to={"activity_type": "course"},
                    related_name="+",
                    to="leprikon.ActivityType",
                    verbose_name="course types",
                ),
            ),
            migrations.AlterField(
                model_name="filteredeventlistplugin",
                name="event_types",
                field=models.ManyToManyField(
                    limit_choices_to={"activity_type": "event"},
                    related_name="+",
                    to="leprikon.ActivityType",
                    verbose_name="event types",
                ),
            ),
            migrations.AlterField(
                model_name="filteredorderablelistplugin",
                name="event_types",
                field=models.ManyToManyField(
                    limit_choices_to={"activity_type": "orderable"},
                    related_name="+",
                    to="leprikon.ActivityType",
                    verbose_name="event types",
                ),
            ),
            migrations.AlterField(
                model_name="orderablelistplugin",
                name="event_types",
                field=models.ManyToManyField(
                    blank=True,
                    help_text="Keep empty to skip searching by event types.",
                    limit_choices_to={"activity_type": "orderable"},
                    related_name="+",
                    to="leprikon.ActivityType",
                    verbose_name="event types",
                ),
            ),
        ]
        + [
            migrations.RenameField(model_name=model_name, old_name="subject_type", new_name="activity_type")
            for model_name in ["registrationlink", "subject", "subjecttypeattachment"]
        ]
        + [
            migrations.RenameModel(old_name="SubjectTypeAttachment", new_name="ActivityTypeAttachment"),
            migrations.RenameModel(old_name="SubjectGroup", new_name="ActivityGroup"),
            migrations.RenameField(model_name="activitygroup", old_name="subject_types", new_name="activity_types"),
        ]
        + [
            migrations.RenameField(model_name=model_name, old_name="subject_ptr", new_name="activity_ptr")
            for model_name in ["course", "event", "orderable"]
        ]
        + [
            leprikon_migrations.AlterModelBases(name=model_name, bases=(models.Model,))
            for model_name in ["course", "event", "orderable"]
        ]
        + [
            migrations.RenameModel(old_name="Subject", new_name="Activity"),
        ]
        + [
            leprikon_migrations.AlterModelBases(name=model_name, bases=("leprikon.activity",))
            for model_name in ["course", "event", "orderable"]
        ]
        + [
            migrations.AlterModelOptions(
                name="activity",
                options={"ordering": ("code", "name"), "verbose_name": "activity", "verbose_name_plural": "activities"},
            ),
            migrations.RenameModel(old_name="SubjectAttachment", new_name="ActivityAttachment"),
            migrations.RenameModel(old_name="SubjectTime", new_name="ActivityTime"),
            migrations.AlterModelOptions(
                name="activitytime",
                options={"ordering": ("days_of_week", "start"), "verbose_name": "time", "verbose_name_plural": "times"},
            ),
            migrations.RenameModel(old_name="SubjectVariant", new_name="ActivityVariant"),
            migrations.AlterModelOptions(
                name="activityvariant",
                options={
                    "ordering": ("activity", "order"),
                    "verbose_name": "price and registering variant",
                    "verbose_name_plural": "price and registering variants",
                },
            ),
            migrations.RenameField(
                model_name="subjectregistration",
                old_name="subject_variant",
                new_name="activity_variant",
            ),
            migrations.RenameModel(old_name="SubjectRegistrationParticipant", new_name="RegistrationParticipant"),
            migrations.RenameModel(old_name="SubjectRegistrationGroup", new_name="RegistrationGroup"),
            migrations.RenameModel(old_name="SubjectRegistrationGroupMember", new_name="RegistrationGroupMember"),
            migrations.RenameModel(old_name="SubjectRegistrationBillingInfo", new_name="RegistrationBillingInfo"),
            migrations.RenameModel(old_name="SubjectPayment", new_name="Payment"),
            migrations.RenameModel(old_name="SubjectReceivedPayment", new_name="ReceivedPayment"),
            migrations.RenameModel(old_name="SubjectReturnedPayment", new_name="ReturnedPayment"),
        ]
        + [
            migrations.RenameField(
                model_name=model_name, old_name="subjectregistration_ptr", new_name="registration_ptr"
            )
            for model_name in ["courseregistration", "eventregistration", "orderableregistration"]
        ]
        + [
            leprikon_migrations.AlterModelBases(name=model_name, bases=(models.Model,))
            for model_name in ["courseregistration", "eventregistration", "orderableregistration"]
        ]
        + [
            migrations.RenameModel(old_name="SubjectRegistration", new_name="Registration"),
        ]
        + [
            leprikon_migrations.AlterModelBases(name=model_name, bases=("leprikon.registration",))
            for model_name in ["courseregistration", "eventregistration", "orderableregistration"]
        ]
        + [
            migrations.RenameField(model_name=model_name, old_name="subject", new_name="activity")
            for model_name in [
                "activityvariant",
                "activityattachment",
                "activitytime",
                "journal",
                "leaderlistplugin",
                "registration",
            ]
        ]
        + [
            migrations.RenameField(
                model_name="registrationlink", old_name="subject_variants", new_name="activity_variants"
            ),
            migrations.AlterModelOptions(
                name="activitygroup",
                options={
                    "ordering": ("order",),
                    "verbose_name": "activity group",
                    "verbose_name_plural": "activity groups",
                },
            ),
            migrations.AlterField(
                model_name="activity",
                name="activity_type",
                field=models.ForeignKey(
                    on_delete=django.db.models.deletion.PROTECT,
                    related_name="activities",
                    to="leprikon.activitytype",
                    verbose_name="activity type",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="age_groups",
                field=models.ManyToManyField(
                    related_name="activities", to="leprikon.AgeGroup", verbose_name="age groups"
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="bill_print_setup",
                field=models.ForeignKey(
                    blank=True,
                    help_text="Only use to set value specific for this activity.",
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name="+",
                    to="leprikon.printsetup",
                    verbose_name="payment print setup",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="decision_print_setup",
                field=models.ForeignKey(
                    blank=True,
                    help_text="Only use to set value specific for this activity.",
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name="+",
                    to="leprikon.printsetup",
                    verbose_name="registration print setup",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="department",
                field=models.ForeignKey(
                    blank=True,
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name="activities",
                    to="leprikon.department",
                    verbose_name="department",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="groups",
                field=models.ManyToManyField(
                    blank=True, related_name="activities", to="leprikon.ActivityGroup", verbose_name="groups"
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="leaders",
                field=models.ManyToManyField(
                    blank=True, related_name="activities", to="leprikon.Leader", verbose_name="leaders"
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="organization",
                field=models.ForeignKey(
                    blank=True,
                    help_text="Only use to set value specific for this activity.",
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name="+",
                    to="leprikon.organization",
                    verbose_name="organization",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="place",
                field=models.ForeignKey(
                    blank=True,
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name="activities",
                    to="leprikon.place",
                    verbose_name="place",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="pr_print_setup",
                field=models.ForeignKey(
                    blank=True,
                    help_text="Only use to set value specific for this activity.",
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name="+",
                    to="leprikon.printsetup",
                    verbose_name="payment request print setup",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="reg_print_setup",
                field=models.ForeignKey(
                    blank=True,
                    help_text="Only use to set value specific for this activity.",
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name="+",
                    to="leprikon.printsetup",
                    verbose_name="registration print setup",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="registration_agreements",
                field=models.ManyToManyField(
                    blank=True,
                    help_text="Add additional legal agreements specific for this activity.",
                    related_name="_leprikon_activity_registration_agreements_+",
                    to="leprikon.Agreement",
                    verbose_name="additional legal agreements",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="school_year",
                field=models.ForeignKey(
                    editable=False,
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name="activities",
                    to="leprikon.schoolyear",
                    verbose_name="school year",
                ),
            ),
            migrations.AlterField(
                model_name="activity",
                name="target_groups",
                field=models.ManyToManyField(
                    related_name="activities", to="leprikon.TargetGroup", verbose_name="target groups"
                ),
            ),
            migrations.AlterField(
                model_name="activityattachment",
                name="activity",
                field=models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name="attachments",
                    to="leprikon.activity",
                    verbose_name="activity",
                ),
            ),
            migrations.AlterField(
                model_name="activitygroup",
                name="activity_types",
                field=models.ManyToManyField(
                    related_name="groups", to="leprikon.ActivityType", verbose_name="activity type"
                ),
            ),
            migrations.AlterField(
                model_name="activitytypeattachment",
                name="activity_type",
                field=models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name="attachments",
                    to="leprikon.activitytype",
                    verbose_name="activity type",
                ),
            ),
            migrations.AlterField(
                model_name="activityvariant",
                name="activity",
                field=models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name="variants",
                    to="leprikon.activity",
                    verbose_name="activity",
                ),
            ),
            migrations.AlterField(
                model_name="activityvariant",
                name="age_groups",
                field=models.ManyToManyField(
                    related_name="activity_variants", to="leprikon.AgeGroup", verbose_name="age groups"
                ),
            ),
            migrations.AlterField(
                model_name="activityvariant",
                name="target_groups",
                field=models.ManyToManyField(
                    related_name="activity_variants", to="leprikon.TargetGroup", verbose_name="target groups"
                ),
            ),
            migrations.AlterField(
                model_name="journal",
                name="activity",
                field=models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name="journals",
                    to="leprikon.activity",
                    verbose_name="activity",
                ),
            ),
            migrations.AlterField(
                model_name="leaderlistplugin",
                name="activity",
                field=models.ForeignKey(
                    blank=True,
                    null=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name="+",
                    to="leprikon.activity",
                    verbose_name="activity",
                ),
            ),
            migrations.AlterField(
                model_name="registration",
                name="activity",
                field=models.ForeignKey(
                    on_delete=django.db.models.deletion.PROTECT,
                    related_name="registrations",
                    to="leprikon.activity",
                    verbose_name="activity",
                ),
            ),
            migrations.AlterField(
                model_name="registrationlink",
                name="activity_variants",
                field=models.ManyToManyField(
                    related_name="registration_links", to="leprikon.ActivityVariant", verbose_name="activities"
                ),
            ),
        ]
    )



from django.db import models
from django.utils.translation import gettext_lazy as _
from djangocms_text_ckeditor.fields import HTMLField


class Place(models.Model):
    name = models.CharField(_("name"), max_length=50)
    place = models.CharField(_("place"), max_length=50, blank=True, default="")
    description = HTMLField(_("description"), blank=True, default="")

    class Meta:
        app_label = "leprikon"
        ordering = ("place",)
        verbose_name = _("place")
        verbose_name_plural = _("places")

    def __str__(self):
        if self.place:
            return "{name}, {place}".format(name=self.name, place=self.place)
        else:
            return "{name}".format(name=self.name)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.27 on 2020-02-26 20:45
from __future__ import unicode_literals

from itertools import chain
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import leprikon.models.fields
from leprikon.utils import get_birth_date


def set_birth_dates(apps, schema_editor):
    Participant = apps.get_model("leprikon", "Participant")
    SubjectRegistrationParticipant = apps.get_model("leprikon", "SubjectRegistrationParticipant")
    for participant in chain(Participant.objects.iterator(), SubjectRegistrationParticipant.objects.iterator()):
        participant.birth_date = get_birth_date(participant.birth_num)
        participant.save()


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0047_pr_print_setup"),
    ]

    operations = [
        migrations.AddField(
            model_name="participant",
            name="birth_date",
            field=models.DateField(default=django.utils.timezone.now, verbose_name="birth date"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="subjectregistrationparticipant",
            name="birth_date",
            field=models.DateField(default=django.utils.timezone.now, verbose_name="birth date"),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="participant",
            name="birth_num",
            field=leprikon.models.fields.BirthNumberField(blank=True, null=True, verbose_name="birth number"),
        ),
        migrations.AlterField(
            model_name="participant",
            name="gender",
            field=models.CharField(
                choices=[("m", "male / boy"), ("f", "female / girl")], max_length=1, verbose_name="gender"
            ),
        ),
        migrations.AlterField(
            model_name="subjectregistrationparticipant",
            name="birth_num",
            field=leprikon.models.fields.BirthNumberField(blank=True, null=True, verbose_name="birth number"),
        ),
        migrations.AlterField(
            model_name="subjectregistrationparticipant",
            name="gender",
            field=models.CharField(
                choices=[("m", "male / boy"), ("f", "female / girl")], max_length=1, verbose_name="gender"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="participant",
            unique_together=set([]),
        ),
        migrations.RunPython(set_birth_dates),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-06-21 13:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0053_registration_dates_by'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='subjectregistration',
            name='cancel_request',
        ),
    ]



from collections import namedtuple
from datetime import date, timedelta
from typing import List

from cms.models import CMSPlugin
from django.db import models, transaction
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from .activities import (
    Activity,
    ActivityDiscount,
    ActivityGroup,
    ActivityModel,
    ActivityType,
    ActivityVariant,
    Registration,
)
from .agegroup import AgeGroup
from .department import Department
from .roles import Leader
from .schoolyear import SchoolYear, SchoolYearDivision, SchoolYearPeriod
from .targetgroup import TargetGroup
from .utils import PaymentStatus, change_year, copy_related_objects


class Course(Activity):
    class Meta:
        app_label = "leprikon"
        ordering = ("code", "name")
        verbose_name = _("course")
        verbose_name_plural = _("courses")

    def copy_to_school_year(old, school_year: SchoolYear):
        new = Course.objects.get(id=old.id)
        new.id, new.pk = None, None
        new.school_year = school_year
        new.public = False
        new.note = ""
        year_delta = school_year.year - old.school_year.year
        new.save()
        new.groups.set(old.groups.all())
        new.age_groups.set(old.age_groups.all())
        new.target_groups.set(old.target_groups.all())
        for leader in old.all_leaders:
            school_year.leaders.add(leader)
        new.leaders.set(old.all_leaders)
        new.questions.set(old.questions.all())
        for old_variant in old.all_variants:
            new_variant = ActivityVariant.objects.get(id=old_variant.id)
            new_variant.id, new_variant.pk = None, None
            new_variant.activity = new
            if old_variant.school_year_division:
                new_variant.school_year_division = SchoolYearDivision.objects.filter(
                    school_year=school_year,
                    name=old_variant.school_year_division.name,
                ).first() or old_variant.school_year_division.copy_to_school_year(school_year)
            new_variant.reg_from = change_year(new_variant.reg_from, year_delta)
            new_variant.reg_to = change_year(new_variant.reg_to, year_delta)
            new_variant.save()
            new_variant.age_groups.set(old_variant.age_groups.all())
            new_variant.target_groups.set(old_variant.target_groups.all())
        copy_related_objects(
            new,
            attachments=old.attachments,
            times=old.times,
        )
        return new


class CourseRegistration(Registration):
    activity_type = ActivityModel.COURSE

    class Meta:
        app_label = "leprikon"
        verbose_name = _("course registration")
        verbose_name_plural = _("course registrations")

    @cached_property
    def all_registration_periods(self) -> List["CourseRegistrationPeriod"]:
        return list(self.course_registration_periods.select_related("period"))

    def get_period_payment_statuses(self, d=None):
        received = self.get_received(d)
        returned = self.get_returned(d)

        for counter, registration_period in enumerate(self.all_registration_periods, start=1):
            period_payment_status = registration_period.get_payment_status(
                received=received,
                returned=returned,
                last_period=counter == len(self.all_registration_periods),
                d=d,
            )
            yield period_payment_status
            received -= period_payment_status.status.received
            returned = 0  # all returned payments are "spent" to the first period
        if not self.all_registration_periods:
            # create fake registration period to ensure some payment status
            registration_period = CourseRegistrationPeriod(
                registration=self,
                period=SchoolYearPeriod(
                    name=_("no period"),
                ),
            )
            yield CourseRegistrationPeriod.PeriodPaymentStatus(
                period=registration_period.period,
                registration_period=registration_period,
                status=PaymentStatus(
                    price=0,
                    discount=0,
                    explanation="",
                    received=received,
                    returned=returned,
                    current_date=d,
                    due_from=None,
                    due_date=None,
                ),
            )

    def get_payment_status(self, d=None):
        payment_status = sum(pps.status for pps in self.get_period_payment_statuses(d))
        if d is None and self.cached_balance != payment_status.balance:
            self.cached_balance = payment_status.balance
            self.save(update_fields=["cached_balance"])
        return payment_status

    @cached_property
    def period_payment_statuses(self):
        return list(self.get_period_payment_statuses())

    @transaction.atomic
    def request_payment(self, payment_requested_by):
        self.course_registration_periods.filter(
            period__due_from__lte=date.today(),
        ).update(payment_requested=True)
        super().request_payment(payment_requested_by)


class CourseRegistrationPeriod(models.Model):
    registration = models.ForeignKey(
        CourseRegistration,
        on_delete=models.CASCADE,
        related_name="course_registration_periods",
        verbose_name=_("registration"),
    )
    period = models.ForeignKey(
        SchoolYearPeriod,
        on_delete=models.PROTECT,
        related_name="course_registration_periods",
        verbose_name=_("period"),
    )
    payment_requested = models.BooleanField(_("payment requested"), default=False)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("registered period")
        verbose_name_plural = _("registered periods")
        ordering = ("period__start",)

    def __str__(self):
        return str(self.period)

    @cached_property
    def all_discounts(self):
        # use cached self.registration.all_discounts instead of
        # return list(self.discounts.all())
        return [discount for discount in self.registration.all_discounts if discount.registration_period_id == self.id]

    class PeriodPaymentStatus(
        namedtuple(
            "PeriodPaymentStatus",
            ["period", "registration_period", "status"],
        )
    ):
        period: SchoolYearPeriod
        registration_period: "CourseRegistrationPeriod"
        status: PaymentStatus

    def get_payment_status(self, received, returned, last_period, d=None) -> PeriodPaymentStatus:
        if d is None:
            d = date.today()
        period_price = self.registration.price * self.period.price_units_count
        discount = sum(discount.amount for discount in self.all_discounts if discount.accounted.date() <= d)
        explanation = ",\n".join(
            discount.explanation.strip()
            for discount in self.all_discounts
            if discount.accounted.date() <= d and discount.explanation.strip()
        )
        return self.PeriodPaymentStatus(
            period=self.period,
            registration_period=self,
            status=PaymentStatus(
                price=period_price,
                discount=discount,
                explanation=explanation,
                received=min(period_price - discount + returned, received) if not last_period else received,
                returned=returned,
                current_date=d,
                due_from=self.registration.payment_requested
                and max(
                    self.period.due_from,
                    self.registration.payment_requested.date(),
                ),
                due_date=self.registration.payment_requested
                and max(
                    self.period.due_date,
                    self.registration.payment_requested.date() + timedelta(days=self.period.min_due_date_days),
                ),
            ),
        )


class CourseDiscount(ActivityDiscount):
    registration = models.ForeignKey(
        CourseRegistration,
        on_delete=models.CASCADE,
        related_name="discounts",
        verbose_name=_("registration"),
    )
    registration_period = models.ForeignKey(
        CourseRegistrationPeriod,
        on_delete=models.CASCADE,
        related_name="discounts",
        verbose_name=_("period"),
    )

    @cached_property
    def period(self):
        return self.registration_period.period

    class Meta:
        app_label = "leprikon"
        verbose_name = _("course discount")
        verbose_name_plural = _("course discounts")
        ordering = ("accounted",)


class CoursePlugin(CMSPlugin):
    course = models.ForeignKey(Course, on_delete=models.CASCADE, related_name="+", verbose_name=_("course"))
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_COURSE_TEMPLATES,
        default=settings.LEPRIKON_COURSE_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"


class CourseListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    departments = models.ManyToManyField(
        Department,
        blank=True,
        related_name="+",
        verbose_name=_("departments"),
        help_text=_("Keep empty to skip searching by departments."),
    )
    course_types = models.ManyToManyField(
        ActivityType,
        blank=True,
        limit_choices_to={"activity_type": ActivityModel.COURSE},
        related_name="+",
        verbose_name=_("course types"),
        help_text=_("Keep empty to skip searching by course types."),
    )
    age_groups = models.ManyToManyField(
        AgeGroup,
        blank=True,
        related_name="+",
        verbose_name=_("age groups"),
        help_text=_("Keep empty to skip searching by age groups."),
    )
    target_groups = models.ManyToManyField(
        TargetGroup,
        blank=True,
        related_name="+",
        verbose_name=_("target groups"),
        help_text=_("Keep empty to skip searching by target groups."),
    )
    groups = models.ManyToManyField(
        ActivityGroup,
        blank=True,
        related_name="+",
        verbose_name=_("course groups"),
        help_text=_("Keep empty to skip searching by groups."),
    )
    leaders = models.ManyToManyField(
        Leader,
        blank=True,
        related_name="+",
        verbose_name=_("leaders"),
        help_text=_("Keep empty to skip searching by leaders."),
    )
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_COURSELIST_TEMPLATES,
        default=settings.LEPRIKON_COURSELIST_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.departments.set(oldinstance.departments.all())
        self.course_types.set(oldinstance.course_types.all())
        self.groups.set(oldinstance.groups.all())
        self.age_groups.set(oldinstance.age_groups.all())
        self.target_groups.set(oldinstance.target_groups.all())
        self.leaders.set(oldinstance.leaders.all())

    @cached_property
    def all_departments(self):
        return list(self.departments.all())

    @cached_property
    def all_course_types(self):
        return list(self.course_types.all())

    @cached_property
    def all_groups(self):
        return list(self.groups.all())

    @cached_property
    def all_age_groups(self):
        return list(self.age_groups.all())

    @cached_property
    def all_target_groups(self):
        return list(self.target_groups.all())

    @cached_property
    def all_leaders(self):
        return list(self.leaders.all())

    Group = namedtuple("Group", ("group", "objects"))

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )
        courses = Course.objects.filter(school_year=school_year, public=True).distinct()

        if self.all_departments:
            courses = courses.filter(department__in=self.all_departments)
        if self.all_course_types:
            courses = courses.filter(activity_type__in=self.all_course_types)
        if self.all_age_groups:
            courses = courses.filter(age_groups__in=self.all_age_groups)
        if self.all_target_groups:
            courses = courses.filter(target_groups__in=self.all_target_groups)
        if self.all_leaders:
            courses = courses.filter(leaders__in=self.all_leaders)
        if self.all_groups:
            courses = courses.filter(groups__in=self.all_groups)
            groups = self.all_groups
        elif self.all_course_types:
            groups = ActivityGroup.objects.filter(activity_types__in=self.all_course_types)
        else:
            groups = ActivityGroup.objects.all()

        context.update(
            {
                "school_year": school_year,
                "courses": courses,
                "groups": (self.Group(group=group, objects=courses.filter(groups=group)) for group in groups),
            }
        )
        return context


class FilteredCourseListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    course_types = models.ManyToManyField(
        ActivityType,
        limit_choices_to={"activity_type": ActivityModel.COURSE},
        related_name="+",
        verbose_name=_("course types"),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.course_types.set(oldinstance.course_types.all())

    @cached_property
    def all_course_types(self):
        return list(self.course_types.all())

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )

        from ..forms.activities import ActivityFilterForm

        form = ActivityFilterForm(
            activity_type_model=ActivityModel.COURSE,
            activity_types=self.all_course_types,
            school_year=school_year,
            is_staff=context["request"].user.is_staff,
            data=context["request"].GET,
        )
        context.update(
            {
                "school_year": school_year,
                "form": form,
                "courses": form.get_queryset(),
            }
        )
        return context



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-10-10 21:11
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0056_registration_period'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='coursediscount',
            name='period',
        ),
        migrations.AlterField(
            model_name='coursediscount',
            name='registration_period',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='discounts', to='leprikon.CourseRegistrationPeriod', verbose_name='period'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-05-06 10:37
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


def set_mail_sent(apps, schema_editor):
    SubjectPayment = apps.get_model('leprikon', 'SubjectPayment')
    SubjectPayment.objects.update(mail_sent=models.F('accounted'))


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0051_registration_group_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='subjectpayment',
            name='mail_sent',
            field=models.DateTimeField(editable=False, null=True, verbose_name='mail sent'),
        ),
        migrations.RunPython(set_mail_sent),
    ]



from collections import namedtuple
from datetime import datetime, timedelta
from typing import Optional

from django.db import models
from django.urls import reverse_lazy as reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from djangocms_text_ckeditor.fields import HTMLField

from ..utils import attributes
from .activities import Activity, RegistrationParticipant
from .pdfmail import PdfExportAndMailMixin
from .roles import Leader
from .schoolyear import SchoolYearDivision, SchoolYearPeriod
from .startend import StartEndMixin
from .times import AbstractTime, TimesMixin


class JournalPeriod:
    def __init__(self, journal: "Journal", period: SchoolYearPeriod = None):
        self.journal: Journal = journal
        self.period: Optional[SchoolYearPeriod] = period

    @property
    def all_journal_entries(self):
        return list(self.journal.journal_entries.filter(period=self.period))

    @cached_property
    def all_participants(self):
        if self.period:  # course
            return [
                participant
                for participant in self.journal.all_participants
                if (
                    (self.period.end is None or participant.created.date() <= self.period.end)
                    and (
                        participant.canceled is None
                        or self.period.start is None
                        or participant.canceled.date() >= self.period.start
                    )
                )
            ]
        else:  # event
            return self.journal.all_participants

    @cached_property
    def all_alternates(self):
        alternates = set()
        for entry in self.all_journal_entries:
            for alternate in entry.all_alternates:
                alternates.add(alternate)
        return list(alternates)

    PresenceRecord = namedtuple("PresenceRecord", ("person", "presences"))

    def get_participant_presences(self):
        return [
            self.PresenceRecord(
                participant, [participant.id in entry.all_participants_idset for entry in self.all_journal_entries]
            )
            for participant in self.all_participants
        ]

    def get_leader_presences(self):
        return [
            self.PresenceRecord(
                leader, [entry.all_leader_entries_by_leader.get(leader, None) for entry in self.all_journal_entries]
            )
            for leader in self.journal.all_leaders
        ]

    def get_alternate_presences(self):
        return [
            self.PresenceRecord(
                alternate,
                [entry.all_leader_entries_by_leader.get(alternate, None) for entry in self.all_journal_entries],
            )
            for alternate in self.all_alternates
        ]


class Journal(PdfExportAndMailMixin, TimesMixin, models.Model):
    object_name = "journal"
    activity = models.ForeignKey(
        Activity, on_delete=models.CASCADE, related_name="journals", verbose_name=_("activity")
    )
    school_year_division = models.ForeignKey(
        SchoolYearDivision,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="journals",
        verbose_name=_("school year division"),
    )
    name = models.CharField(_("journal name"), blank=True, default="", max_length=150)
    leaders = models.ManyToManyField(Leader, blank=True, related_name="journals", verbose_name=_("leaders"))
    participants = models.ManyToManyField(
        RegistrationParticipant, blank=True, related_name="journals", verbose_name=_("participants")
    )
    risks = HTMLField(_("risks"), blank=True)
    plan = HTMLField(_("plan"), blank=True)
    evaluation = HTMLField(_("evaluation"), blank=True)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("journal")
        verbose_name_plural = _("journal")

    def __str__(self):
        return f"{self.activity.display_name} - {self.name}" if self.name else self.activity.display_name

    @cached_property
    def all_journal_entries(self):
        return list(self.journal_entries.all())

    @cached_property
    def all_journal_periods(self):
        if self.school_year_division:
            return [JournalPeriod(self, period) for period in self.school_year_division.all_periods]
        return [JournalPeriod(self)]

    @cached_property
    def all_leaders(self):
        return list(self.leaders.all())

    @cached_property
    def all_participants(self):
        return list(
            self.participants.annotate(
                created=models.F("registration__created"),
                approved=models.F("registration__approved"),
                canceled=models.F("registration__canceled"),
            )
        )

    def get_absolute_url(self):
        return reverse("leprikon:journal", kwargs={"pk": self.pk})

    def get_valid_participants(self, d):
        return self.participants.exclude(registration__canceled__date__lt=d)


class JournalTime(AbstractTime):
    journal = models.ForeignKey(Journal, on_delete=models.CASCADE, related_name="times", verbose_name=_("journal"))

    class Meta:
        app_label = "leprikon"
        ordering = ("start_date", "days_of_week", "start_time")
        verbose_name = _("time")
        verbose_name_plural = _("times")


class JournalEntry(StartEndMixin, models.Model):
    journal = models.ForeignKey(
        Journal, editable=False, on_delete=models.PROTECT, related_name="journal_entries", verbose_name=_("journal")
    )
    period = models.ForeignKey(
        SchoolYearPeriod,
        null=True,
        on_delete=models.PROTECT,
        related_name="journal_entries",
        verbose_name=_("school year period"),
    )
    date = models.DateField(_("date"))
    start = models.TimeField(_("start time"), blank=True, null=True)
    end = models.TimeField(_("end time"), blank=True, null=True)
    agenda = HTMLField(_("session agenda"), default="")
    participants = models.ManyToManyField(
        RegistrationParticipant, blank=True, related_name="journal_entries", verbose_name=_("participants")
    )
    participants_instructed = models.ManyToManyField(
        RegistrationParticipant,
        blank=True,
        related_name="instructed",
        verbose_name=_("participants instructed about safety and internal rules"),
    )

    class Meta:
        app_label = "leprikon"
        ordering = ("date", "start", "end")
        verbose_name = _("journal entry")
        verbose_name_plural = _("journal entries")

    def __str__(self):
        return f"{self.journal}, {self.date}, {self.duration}"

    @cached_property
    def datetime_start(self):
        try:
            return datetime.combine(self.date, self.start)
        except TypeError:
            return None

    @cached_property
    def datetime_end(self):
        try:
            return datetime.combine(self.date, self.end)
        except TypeError:
            return None

    @attributes(short_description=_("duration"))
    @cached_property
    def duration(self):
        try:
            return self.datetime_end - self.datetime_start
        except TypeError:
            return timedelta()

    @cached_property
    def all_participants(self):
        return list(self.participants.all())

    @cached_property
    def all_participants_instructed(self):
        return list(self.participants_instructed.all())

    @cached_property
    def all_participants_idset(self):
        return set(r.id for r in self.all_participants)

    @cached_property
    def all_leader_entries(self):
        return list(self.leader_entries.all())

    @cached_property
    def all_leader_entries_by_leader(self):
        return dict((e.timesheet.leader, e) for e in self.all_leader_entries)

    @cached_property
    def all_leaders(self):
        return list(
            le.timesheet.leader for le in self.all_leader_entries if le.timesheet.leader in self.journal.all_leaders
        )

    @cached_property
    def all_alternates(self):
        return list(
            le.timesheet.leader for le in self.all_leader_entries if le.timesheet.leader not in self.journal.all_leaders
        )

    @cached_property
    def affects_submitted_timesheets(self):
        return self.leader_entries.filter(timesheet__submitted=True).exists()

    def save(self, *args, **kwargs):
        if self.end is None:
            self.end = self.start
        super().save(*args, **kwargs)

    def get_edit_url(self):
        return reverse("leprikon:journalentry_update", args=(self.id,))

    def get_delete_url(self):
        return reverse("leprikon:journalentry_delete", args=(self.id,))


class JournalLeaderEntry(StartEndMixin, models.Model):
    journal_entry = models.ForeignKey(
        JournalEntry,
        editable=False,
        on_delete=models.CASCADE,
        related_name="leader_entries",
        verbose_name=_("journal entry"),
    )
    timesheet = models.ForeignKey(
        "leprikon.Timesheet",
        editable=False,
        on_delete=models.PROTECT,
        related_name="journal_entries",
        verbose_name=_("timesheet"),
    )
    start = models.TimeField(_("start time"))
    end = models.TimeField(_("end time"))

    class Meta:
        app_label = "leprikon"
        verbose_name = _("journal leader entry")
        verbose_name_plural = _("journal leader entries")
        unique_together = (("journal_entry", "timesheet"),)

    def __str__(self):
        return f"{self.journal_entry}"

    @attributes(admin_order_field="journal_entry__date", short_description=_("date"))
    @cached_property
    def date(self):
        return self.journal_entry.date

    @attributes(short_description=_("journal"))
    @cached_property
    def journal(self):
        return self.journal_entry.journal

    @attributes(short_description=_("activity"))
    @cached_property
    def activity(self):
        return self.journal_entry.journal.activity

    @cached_property
    def datetime_start(self):
        return datetime.combine(self.date, self.start)

    @cached_property
    def datetime_end(self):
        return datetime.combine(self.date, self.end)

    @attributes(short_description=_("duration"))
    @cached_property
    def duration(self):
        return self.datetime_end - self.datetime_start

    @property
    def group(self):
        return self.activity

    def get_edit_url(self):
        return reverse("leprikon:journalleaderentry_update", args=(self.id,))

    def get_delete_url(self):
        return reverse("leprikon:journalleaderentry_delete", args=(self.id,))



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class LeprikonConfig(AppConfig):
    name = "leprikon"
    verbose_name = _("Leprikon")

    def ready(self):
        # register bank statement readers
        from . import bankreaders  # noqa

        # ensure that current LeprikonSite exists
        from .models.leprikonsite import LeprikonSite

        try:
            LeprikonSite.objects.get_current()
        except Exception:
            pass

        # create leprikon page on first run
        from cms.api import create_page
        from cms.constants import TEMPLATE_INHERITANCE_MAGIC
        from menus.menu_pool import menu_pool

        from .conf import settings

        try:
            create_page(
                title="Leprikón",
                template=TEMPLATE_INHERITANCE_MAGIC,
                language=settings.LANGUAGE_CODE,
                slug="leprikon",
                apphook="LeprikonApp",
                apphook_namespace="leprikon",
                reverse_id="leprikon",
                in_navigation=True,
                navigation_extenders="LeprikonMenu",
                published=True,
            ).set_as_homepage()
            menu_pool.clear()
        except Exception:
            pass



from adminsortable2.admin import SortableInlineAdminMixin
from django import forms
from django.contrib import admin
from django.db import IntegrityError
from django.shortcuts import render
from django.utils.translation import gettext_lazy as _

from ..models.schoolyear import SchoolYear, SchoolYearDivision, SchoolYearPeriod
from ..utils import attributes
from .filters import SchoolYearListFilter


@admin.register(SchoolYear)
class SchoolYearAdmin(admin.ModelAdmin):
    list_display = ("name", "active")
    list_editable = ("active",)

    # do not allow to delete entries in admin
    def has_delete_permission(self, request, obj=None):
        return False

    def get_readonly_fields(self, request, obj=None):
        return obj and ("year",) or ()

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions


class SchoolYearPeriodInlineAdmin(SortableInlineAdminMixin, admin.TabularInline):
    model = SchoolYearPeriod
    extra = 0
    min_num = 1


@admin.register(SchoolYearDivision)
class SchoolYearDivisionAdmin(admin.ModelAdmin):
    actions = ("copy_to_school_year",)
    inlines = (SchoolYearPeriodInlineAdmin,)
    list_display = ("name",)
    list_filter = (("school_year", SchoolYearListFilter),)

    @attributes(short_description=_("Copy selected school year divisions to another school year"))
    def copy_to_school_year(self, request, queryset):
        class SchoolYearForm(forms.Form):
            school_year = forms.ModelChoiceField(
                label=_("Target school year"),
                help_text=_("All selected school year divisions will be copied to selected school year."),
                queryset=SchoolYear.objects.all(),
            )

        if request.POST.get("post", "no") == "yes":
            form = SchoolYearForm(request.POST)
            if form.is_valid():
                school_year = form.cleaned_data["school_year"]
                for school_year_division in queryset.all():
                    try:
                        school_year_division.copy_to_school_year(school_year)
                    except IntegrityError:
                        # division already exists for target year
                        pass
                self.message_user(
                    request,
                    _("Selected school year divisions were copied to school year {}.").format(school_year),
                )
                return
        else:
            form = SchoolYearForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {"fields": list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(
            request,
            "leprikon/admin/action_form.html",
            dict(
                title=_("Select target school year"),
                opts=self.model._meta,
                adminform=adminform,
                media=self.media + adminform.media,
                action="copy_to_school_year",
                action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
                select_across=request.POST["select_across"],
                selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
            ),
        )

    def get_readonly_fields(self, request, obj=None):
        return ("school_year",) if obj else ()



from time import time

from django.contrib.sites.models import Site
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django_pays.models import Gateway
from djangocms_text_ckeditor.fields import HTMLField

from ..conf import settings
from .account import AccountClosure
from .agreements import Agreement
from .organizations import Organization
from .printsetup import PrintSetup


class LeprikonSiteManager(models.Manager):
    _cached_site = None
    _cache_timestamp = None

    def get_current(self):
        now = time()
        if not self._cached_site or now - self._cache_timestamp > settings.LEPRIKON_SITE_CACHE_MAX_AGE:
            lookup_kwargs = {}
            if getattr(settings, "SITE_ID", ""):
                lookup_kwargs["pk"] = settings.SITE_ID
            self._cached_site = self.get_or_create(**lookup_kwargs)[0]
            self._cache_timestamp = now
        return self._cached_site

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related(
                "bill_print_setup",
                "reg_print_setup",
            )
        )


class LeprikonSite(Site):
    organization = models.ForeignKey(
        Organization,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("default organization"),
    )
    reg_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("default registration print setup"),
    )
    decision_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("default decision print setup"),
    )
    pr_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("default payment request print setup"),
    )
    bill_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("default bill print setup"),
    )
    user_agreement = HTMLField(_("user agreement"), blank=True, default="")
    user_agreement_changed = models.DateTimeField(_("last time user agreement changed"), blank=True, null=True)
    registration_agreements = models.ManyToManyField(
        Agreement,
        blank=True,
        related_name="+",
        verbose_name=_("default registration agreements"),
        help_text=_("Add legal agreements for the registration form."),
    )
    payment_gateway = models.ForeignKey(
        Gateway, blank=True, null=True, on_delete=models.SET_NULL, related_name="+", verbose_name=_("Payment Gateway")
    )

    objects = LeprikonSiteManager()

    class Meta:
        app_label = "leprikon"
        verbose_name = _("leprikon site")
        verbose_name_plural = _("leprikon sites")

    @cached_property
    def max_closure_date(self):
        return AccountClosure.objects.max_closure_date()

    @cached_property
    def url(self):
        return settings.LEPRIKON_URL or f"https://{self.domain}"



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2021-06-01 01:27
from __future__ import unicode_literals

from decimal import Decimal
from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


def update_subject_payments(apps, schema_editor):
    SubjectPayment = apps.get_model("leprikon", "SubjectPayment")

    # delete zero amount payments just to be sure
    SubjectPayment.objects.filter(related_payment__amount=Decimal("0")).update(related_payment=None)
    SubjectPayment.objects.filter(amount=Decimal("0")).delete()

    # fix inconsistent payment types :rolling eyes:
    for wrong, correct in (
        ("RETURN_CASH", "PAYMENT_CASH"),
        ("RETURN_BANK", "PAYMENT_BANK"),
        ("RETURN_TRANSFER", "PAYMENT_TRANSFER"),
    ):
        SubjectPayment.objects.filter(amount__gt=Decimal("0"), payment_type=wrong).update(payment_type=correct)
    for wrong, correct in (
        ("PAYMENT_CASH", "RETURN_CASH"),
        ("PAYMENT_BANK", "RETURN_BANK"),
        ("PAYMENT_ONLINE", "RETURN_BANK"),
        ("PAYMENT_TRANSFER", "RETURN_TRANSFER"),
    ):
        SubjectPayment.objects.filter(amount__lt=Decimal("0"), payment_type=wrong).update(payment_type=correct)

    # set target registration on payments
    SubjectPayment.objects.filter(amount__gt=Decimal("0")).update(
        target_registration=models.F("registration"),
    )

    # set source registration and amount on returned payments
    SubjectPayment.objects.filter(amount__lt=Decimal("0")).update(
        source_registration=models.F("registration"),
        amount=0 - models.F("amount"),
    )

    # set source registration on transfer payments and remove related returned payments
    for payment in SubjectPayment.objects.filter(
        payment_type="PAYMENT_TRANSFER",
        related_payment__payment_type="RETURN_TRANSFER",
        related_payment__amount=models.F("amount"),
    ).select_related("related_payment"):
        payment.source_registration_id = payment.related_payment.registration_id
        payment.note = f"{payment.note}, {payment.related_payment.note}".strip(", ")
        payment.save()
        SubjectPayment.objects.filter(related_payment=payment.related_payment).update(related_payment=None)
        payment.related_payment.delete()
    # also in other direction
    for payment in SubjectPayment.objects.filter(
        payment_type="RETURN_TRANSFER",
        related_payment__payment_type="PAYMENT_TRANSFER",
        related_payment__amount=models.F("amount"),
    ).select_related("related_payment"):
        payment.related_payment.source_registration_id = payment.registration_id
        payment.related_payment.note = f"{payment.related_payment.note}, {payment.note}".strip(", ")
        payment.related_payment.save()
        SubjectPayment.objects.filter(related_payment=payment).update(related_payment=None)
        payment.delete()

    # update obsolete payment types
    SubjectPayment.objects.filter(
        payment_type__in=("PAYMENT_TRANSFER", "RETURN_TRANSFER"),
    ).update(payment_type="TRANSFER")

    # find related payments
    for payment in SubjectPayment.objects.filter(
        payment_type="TRANSFER",
        source_registration=None,
    ).select_related("target_registration"):
        related_payment = SubjectPayment.objects.filter(
            payment_type="TRANSFER",
            target_registration=None,
            amount=payment.amount,
            source_registration__user_id=payment.target_registration.user_id,
            accounted__date=payment.accounted.date(),
        ).first()
        if related_payment:
            payment.source_registration_id = related_payment.source_registration_id
            payment.save()
            related_payment.delete()


class Migration(migrations.Migration):

    dependencies = [
        ("django_pays", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("bankreader", "0003_transaction_account_statement_cleanup"),
        ("leprikon", "0058_agegroup_min_max_age"),
    ]

    operations = [
        migrations.AddField(
            model_name="subjectpayment",
            name="source_registration",
            field=models.ForeignKey(
                blank=True,
                help_text="The payment will be deducted from this registration.",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="returned_payments",
                to="leprikon.SubjectRegistration",
                verbose_name="source registration",
            ),
        ),
        migrations.AddField(
            model_name="subjectpayment",
            name="target_registration",
            field=models.ForeignKey(
                blank=True,
                help_text="The payment will be added to this registration.",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="received_payments",
                to="leprikon.SubjectRegistration",
                verbose_name="target registration",
            ),
        ),
        migrations.RunPython(update_subject_payments),
    ]



from django.contrib import messages
from django.urls import reverse_lazy as reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views.generic import (
    CreateView as _CreateView,
    DeleteView as _DeleteView,
    DetailView as _DetailView,
    FormView as _FormView,
    ListView as _ListView,
    TemplateView as _TemplateView,
    UpdateView as _UpdateView,
)

from ..conf import settings
from ..forms.confirm import ConfirmForm


class GenericViewMixin(object):
    title = ""

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, placeholder=self.get_placeholder(), title=self.get_title(), **kwargs)
        return context

    def get_placeholder(self):
        return self.__class__.__name__

    def get_title(self):
        return self.title


class ListView(GenericViewMixin, _ListView):
    template_name = "leprikon/list.html"
    add_url = None
    add_label = _("add")
    add_title = ""
    message_empty = _("No items found.")
    preview_template = ""

    def get_add_url(self):
        return self.add_url

    def get_add_label(self):
        return self.add_label

    def get_add_title(self):
        return self.add_title

    def get_message_empty(self):
        return self.message_empty

    def get_context_data(self, *args, **kwargs):
        return super().get_context_data(
            *args,
            preview_template=self.preview_template,
            add_url=self.get_add_url(),
            add_label=self.get_add_label(),
            add_title=self.get_add_title(),
            message_empty=self.get_message_empty(),
            **kwargs
        )


class FilteredListView(ListView):
    form_class = None
    message_empty = _("No items found matching given search parameters.")

    @cached_property
    def form(self):
        return self.get_form()

    def get_form(self):
        return self.form_class(data=self.request.GET)

    def get_context_data(self, *args, **kwargs):
        return super().get_context_data(*args, form=self.form, **kwargs)

    def get_queryset(self):
        return self.form.get_queryset()


class BackViewMixin(object):
    back_url = reverse("leprikon:summary")
    back_label = _("Back")

    def get_context_data(self, *args, **kwargs):
        return super().get_context_data(*args, back_url=self.get_back_url(), back_label=self.get_back_label(), **kwargs)

    def get_back_label(self):
        return self.back_label

    def get_back_url(self):
        url = self.request.POST.get(
            settings.LEPRIKON_PARAM_BACK, self.request.GET.get(settings.LEPRIKON_PARAM_BACK, "")
        )
        if url.startswith("/"):
            return url
        else:
            return self.back_url


class FormViewMixin(BackViewMixin, GenericViewMixin):
    template_name = "leprikon/form.html"
    form_item_template_name = "leprikon/form_item.html"
    instructions = ""
    submit_label = _("Save")
    success_url = reverse("leprikon:summary")
    message = None

    def get_context_data(self, *args, **kwargs):
        return super().get_context_data(
            *args,
            form_item_template=self.get_form_item_template_name(),
            instructions=self.get_instructions(),
            submit_label=self.get_submit_label(),
            **kwargs
        )

    def get_form_item_template_name(self):
        return self.form_item_template_name

    def get_instructions(self):
        return self.instructions

    def get_submit_label(self):
        return self.submit_label

    def get_success_url(self):
        return self.get_back_url()

    def get_message(self):
        return self.message

    def form_valid(self, form):
        response = super().form_valid(form)
        message = self.get_message()
        if message:
            messages.info(
                self.request,
                message,
            )
        return response


class ConfirmFormViewMixin(FormViewMixin):
    form_class = ConfirmForm
    template_name = "leprikon/confirm_form.html"
    question = ""
    submit_label = _("Yes")

    def get_question(self):
        return self.question

    def get_context_data(self, *args, **kwargs):
        return super().get_context_data(*args, question=self.get_question(), **kwargs)

    def form_valid(self, form):
        self.confirmed()
        return super().form_valid(form)

    def confirmed(self):
        pass


class DetailView(GenericViewMixin, _DetailView):
    pass


class CreateView(FormViewMixin, _CreateView):
    pass


class ConfirmCreateView(ConfirmFormViewMixin, _CreateView):
    pass


class UpdateView(FormViewMixin, _UpdateView):
    pass


class ConfirmUpdateView(ConfirmFormViewMixin, _UpdateView):
    pass


class DeleteView(ConfirmFormViewMixin, _DeleteView):
    submit_label = _("Delete")

    def delete(self, request, *args, **kwargs):
        response = super().delete(request, *args, **kwargs)
        message = self.get_message()
        if message:
            messages.info(
                self.request,
                message,
            )
        return response


class FormView(FormViewMixin, _FormView):
    pass


class TemplateView(GenericViewMixin, _TemplateView):
    pass



from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from itertools import chain, product
from typing import Iterable, Iterator, Optional

from dateutil.rrule import DAILY, FR, MO, SA, SU, TH, TU, WE, rrule, weekday
from django.db.models import IntegerChoices
from django.utils.translation import gettext_lazy as _

from . import comma_separated


def start_time_format(start: time) -> str:
    return start.strftime("%H:%M")


def end_time_format(end: time) -> str:
    return "24:00" if end == time(0) else end.strftime("%H:%M")


def time_slot_format(start_time: time, end_time: time) -> str:
    return _("{start_time} - {end_time}").format(
        start_time=start_time_format(start_time), end_time=end_time_format(end_time)
    )


class DayOfWeek(IntegerChoices):
    MONDAY = 1 << 0, _("Monday")
    TUESDAY = 1 << 1, _("Tuesday")
    WEDNESDAY = 1 << 2, _("Wednesday")
    THURSDAY = 1 << 3, _("Thursday")
    FRIDAY = 1 << 4, _("Friday")
    SATURDAY = 1 << 5, _("Saturday")
    SUNDAY = 1 << 6, _("Sunday")

    def isoweekday(self) -> int:
        """
        Returns the ISO weekday number (1=Monday, 7=Sunday).
        """
        return {
            self.MONDAY: 1,
            self.TUESDAY: 2,
            self.WEDNESDAY: 3,
            self.THURSDAY: 4,
            self.FRIDAY: 5,
            self.SATURDAY: 6,
            self.SUNDAY: 7,
        }[
            self  # type: ignore
        ]

    @classmethod
    def from_isoweekday(cls, iso_weekday: int) -> "DayOfWeek":
        """
        Returns the DayOfWeek from an ISO weekday number (1=Monday, 7=Sunday).
        """
        return [
            cls.SUNDAY,
            cls.MONDAY,
            cls.TUESDAY,
            cls.WEDNESDAY,
            cls.THURSDAY,
            cls.FRIDAY,
            cls.SATURDAY,
            cls.SUNDAY,
        ][iso_weekday]


class DaysOfWeek(list[DayOfWeek]):

    def __init__(self, value: int | Iterable[DayOfWeek] = 0) -> None:
        if isinstance(value, (int, DayOfWeek)):
            return super().__init__(day for day in DayOfWeek if value & day)
        return super().__init__(DayOfWeek(day) for day in value)

    def int(self) -> int:
        return sum(int(i) for i in self)

    def __str__(self) -> str:
        i = self.int()
        parts = []
        sequence = []
        for day in DayOfWeek:
            available = day & i
            if available:
                sequence.append(day.label)
            if (not available or day == DayOfWeek.SUNDAY) and sequence:
                if len(sequence) == 1:
                    parts.append(sequence[0])
                else:
                    parts.append(f"{sequence[0]} - {sequence[-1]}")
                sequence = []
        return comma_separated(parts)

    def __and__(self, other: "DaysOfWeek") -> "DaysOfWeek":
        return DaysOfWeek(self.int() & other.int())

    @classmethod
    def all(cls) -> "DaysOfWeek":
        """
        Returns all days of the week.
        """
        return DaysOfWeek(DayOfWeek)


@dataclass
class WeeklyTime:
    """
    Represents a weekly time range.
    If end_time is 0:00, it actually means 24:00 (end of the day)
    """

    start_date: Optional[date]
    end_date: Optional[date]
    days_of_week: DaysOfWeek
    start_time: time
    end_time: time

    def __str__(self):
        return _("{days}, {time}").format(
            days=self.days_of_week,
            time=time_slot_format(self.start_time, self.end_time),
        )

    def __bool__(self) -> bool:
        if not self.days_of_week:
            return False
        if self.start_time >= self.end_time and self.end_time != time(0):
            return False
        if self.start_date and self.end_date and self.start_date > self.end_date:
            return False
        return True

    def __and__(self, other: "WeeklyTime") -> Optional["WeeklyTime"]:
        return (
            WeeklyTime(
                start_date=(
                    max(self.start_date, other.start_date)
                    if self.start_date and other.start_date
                    else self.start_date or other.start_date
                ),
                end_date=(
                    min(self.end_date, other.end_date)
                    if self.end_date and other.end_date
                    else self.end_date or other.end_date
                ),
                days_of_week=DaysOfWeek(self.days_of_week & other.days_of_week),
                start_time=max(self.start_time, other.start_time),
                # if any of the end times is 0:00:00 (which means EOD), use the other one
                end_time=(
                    max(self.end_time, other.end_time)
                    if time(0) in (self.end_time, other.end_time)
                    else min(self.end_time, other.end_time)
                ),
            )
            or None
        )

    @classmethod
    def unlimited(cls) -> "WeeklyTime":
        return WeeklyTime(None, None, DaysOfWeek.all(), time(0), time(0))


class WeeklyTimes(list[WeeklyTime]):
    def __and__(self, other: "WeeklyTimes") -> "WeeklyTimes":
        return WeeklyTimes(a for a in [a1 & a2 for a1, a2 in product(self, other)] if a)

    def __str__(self):
        return comma_separated(self)

    @classmethod
    def unlimited(cls) -> "WeeklyTimes":
        return WeeklyTimes([WeeklyTime.unlimited()])


@dataclass
class TimeSlot:
    start: datetime
    end: datetime

    def __hash__(self) -> int:
        return hash((self.start, self.end))

    def __post_init__(self):
        assert self.start < self.end

    def __and__(self, other: "TimeSlot") -> "TimeSlots":
        if self.start < other.end and self.end > other.start:
            return TimeSlots([TimeSlot(max(self.start, other.start), min(self.end, other.end))])
        return TimeSlots()

    def __or__(self, other: "TimeSlot") -> "TimeSlots":
        return TimeSlots([self, other])

    def __sub__(self, other: "TimeSlot | TimeSlots") -> "TimeSlots":
        if isinstance(other, TimeSlot):
            if self.start >= other.end or self.end <= other.start:
                return TimeSlots([self])
            elif self.start < other.start:
                if self.end > other.end:
                    return TimeSlots(
                        [
                            TimeSlot(self.start, other.start),
                            TimeSlot(other.end, self.end),
                        ]
                    )
                else:
                    return TimeSlots([TimeSlot(self.start, other.start)])
            else:
                if self.end > other.end:
                    return TimeSlots([TimeSlot(other.end, self.end)])
                else:
                    return TimeSlots()
        return TimeSlots([self]) - other


class TimeSlots(list[TimeSlot]):
    def __init__(self, time_slots: Iterable[TimeSlot] = []):
        normalized: list[TimeSlot] = []
        for ts in sorted(time_slots, key=lambda ts: (ts.start, ts.end)):
            if not normalized or normalized[-1].end < ts.start:
                normalized.append(ts)
            else:
                normalized[-1].end = max(normalized[-1].end, ts.end)
        super().__init__(normalized)

    def __and__(self, other: "TimeSlots") -> "TimeSlots":
        return TimeSlots(chain.from_iterable(a1 & a2 for a1, a2 in product(self, other)))

    def __or__(self, other: "TimeSlots") -> "TimeSlots":
        return TimeSlots(self + other)

    def __sub__(self, other: "TimeSlot | TimeSlots") -> "TimeSlots":
        if isinstance(other, TimeSlot):
            return TimeSlots(chain.from_iterable(ts - other for ts in self))
        result = TimeSlots(self)
        for ts in other:
            result = result - ts
        return result


@dataclass
class SimpleEvent:
    timeslot: TimeSlot
    resource_groups: list[set[int]]

    def has_resolvable_resource_groups(self) -> bool:
        number_of_resources = len(self.resource_groups)
        return any(len(set(resources)) == number_of_resources for resources in product(*self.resource_groups))


def get_byweekdays_by_days_of_week(days_of_week: DaysOfWeek) -> list[weekday]:
    return [
        {
            DayOfWeek.MONDAY: MO,
            DayOfWeek.TUESDAY: TU,
            DayOfWeek.WEDNESDAY: WE,
            DayOfWeek.THURSDAY: TH,
            DayOfWeek.FRIDAY: FR,
            DayOfWeek.SATURDAY: SA,
            DayOfWeek.SUNDAY: SU,
        }[day]
        for day in days_of_week
    ]


def get_time_slots_by_weekly_times(
    week_times: WeeklyTimes,
    start_date: date,
    end_date: date,
) -> TimeSlots:
    return TimeSlots(
        TimeSlot(
            start=datetime.combine(dt.date(), week_time.start_time),
            end=(
                datetime.combine(dt.date(), week_time.end_time)
                if week_time.end_time != time(0)
                else dt + timedelta(days=1)
            ),
        )
        for week_time in week_times
        for dt in rrule(
            DAILY,
            dtstart=max(week_time.start_date, start_date) if week_time.start_date else start_date,
            until=min(week_time.end_date, end_date) if week_time.end_date else end_date,
            byweekday=get_byweekdays_by_days_of_week(week_time.days_of_week),
        )
    )


def get_unavailable_time_slots(
    available_time_slots: TimeSlots,
    start_date: date,
    end_date: date,
) -> TimeSlots:
    return (
        TimeSlot(
            start=datetime.combine(start_date, time(0)),
            end=datetime.combine(end_date, time(0)) + timedelta(days=1),
        )
        - available_time_slots
    )


def add_event_to_flattened_events(flattened_events: Iterable[SimpleEvent], event: SimpleEvent) -> Iterator[SimpleEvent]:
    """Add an event to an ordered list of non overlaping events."""
    new_event: SimpleEvent | None = event
    for current_event in flattened_events:
        # nothing to merge
        if new_event is None:
            yield current_event
            continue

        # current event before new event
        if current_event.timeslot.end <= new_event.timeslot.start:
            yield current_event
            continue

        # current event after new event
        if current_event.timeslot.start >= new_event.timeslot.end:
            yield new_event
            new_event = None
            yield current_event
            continue

        # first non overlaping part
        if current_event.timeslot.start < new_event.timeslot.start:
            yield SimpleEvent(
                timeslot=TimeSlot(current_event.timeslot.start, new_event.timeslot.start),
                resource_groups=current_event.resource_groups,
            )
        elif new_event.timeslot.start < current_event.timeslot.start:
            yield SimpleEvent(
                timeslot=TimeSlot(new_event.timeslot.start, current_event.timeslot.start),
                resource_groups=new_event.resource_groups,
            )

        # overlaping part
        overlaping_start = max(new_event.timeslot.start, current_event.timeslot.start)
        overlaping_end = min(new_event.timeslot.end, current_event.timeslot.end)
        yield SimpleEvent(
            timeslot=TimeSlot(overlaping_start, overlaping_end),
            resource_groups=current_event.resource_groups + new_event.resource_groups,
        )

        # remaining part of current event
        if current_event.timeslot.end > overlaping_end:
            yield SimpleEvent(
                timeslot=TimeSlot(overlaping_end, current_event.timeslot.end),
                resource_groups=current_event.resource_groups,
            )

        # remaining part of new event
        if new_event.timeslot.end > overlaping_end:
            new_event = SimpleEvent(
                timeslot=TimeSlot(overlaping_end, new_event.timeslot.end),
                resource_groups=new_event.resource_groups,
            )
        else:
            new_event = None

    if new_event:
        yield new_event


def flatten_events(events: Iterable[SimpleEvent]) -> Iterator[SimpleEvent]:
    """Turn possibly overlaping events into non overlaping events by merging them."""
    flattened_events: Iterator[SimpleEvent] = iter(())
    for event in events:
        flattened_events = add_event_to_flattened_events(flattened_events, event)
    return flattened_events


def get_conflicting_timeslots(events: Iterable[SimpleEvent]) -> Iterator[TimeSlot]:
    for event in flatten_events(events):
        if not event.has_resolvable_resource_groups():
            yield event.timeslot


def apply_preparation_and_recovery_times(
    conflicting_time_slots: Iterable[TimeSlot], preparation_time: timedelta, recovery_time: timedelta
) -> TimeSlots:
    """Apply preparation and recovery time to time conflicting times slots.
    Preparation time is applied to the ends of the conflicting timeslots, and vice versa."""
    return TimeSlots(
        TimeSlot(
            start=ts.start - recovery_time,
            end=ts.end + preparation_time,
        )
        for ts in conflicting_time_slots
    )



from typing import Any, List, Optional

from adminsortable2.admin import SortableAdminMixin, SortableInlineAdminMixin
from django import forms
from django.contrib import admin, messages
from django.contrib.admin.templatetags.admin_list import _boolean_icon
from django.contrib.admin.utils import unquote
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db.models import BooleanField, F, Func
from django.db.models.functions import Coalesce, Random
from django.http import HttpRequest, HttpResponseRedirect
from django.template.response import SimpleTemplateResponse
from django.templatetags.static import static
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ..forms.activities import (
    ActivityAdminForm,
    ActivityVariantForm,
    RegistrationAdminForm,
    RegistrationGroupAdminForm,
    RegistrationParticipantAdminForm,
)
from ..models.activities import (
    DEFAULT_TEXTS,
    Activity,
    ActivityAttachment,
    ActivityGroup,
    ActivityModel,
    ActivityTime,
    ActivityType,
    ActivityTypeAttachment,
    ActivityTypePrintSetups,
    ActivityTypeTexts,
    ActivityVariant,
    ReceivedPayment,
    Registration,
    RegistrationBillingInfo,
    RegistrationGroup,
    RegistrationGroupMember,
    RegistrationParticipant,
    ReturnedPayment,
)
from ..models.utils import lazy_help_text_with_html_default
from ..utils import amount_color, attributes, currency
from .bulkupdate import BulkUpdateMixin
from .export import AdminExportMixin
from .filters import (
    ActivityGroupListFilter,
    ActivityListFilter,
    ActivityTypeListFilter,
    ApprovedListFilter,
    CanceledListFilter,
    IsNullFieldListFilter,
    LeaderListFilter,
    SchoolYearListFilter,
)
from .messages import SendMessageAdminMixin
from .sendmail import SendMailAdminMixin
from .settings import ChildModelAdmin, ParentAdminMixin
from .transaction import TransactionAdminMixin, TransactionBaseAdmin, TransactionTypeListFilter
from .utils import datetime_with_by


class IsNull(Func):
    _output_field = BooleanField()
    arity = 1
    template = "%(expressions)s IS NULL"


class ActivityTypeAttachmentInlineAdmin(SortableInlineAdminMixin, admin.TabularInline):
    model = ActivityTypeAttachment
    extra = 0


ACTIVITY_TYPE_TEXT_FIELDS = [f.name for f in ActivityTypeTexts._meta.fields if f.name.startswith("text_")]
ACTIVITY_TYPE_PRINT_SETUP_FIELDS = [
    f.name for f in ActivityTypePrintSetups._meta.fields if f.name.endswith("_print_setup")
]


@admin.register(ActivityType)
class ActivityTypeAdmin(ParentAdminMixin, BulkUpdateMixin, SortableAdminMixin, admin.ModelAdmin):
    child_models = [ActivityTypeTexts, ActivityTypePrintSetups]
    bulk_update_exclude = (
        "model",
        "page",
        "slug",
        "order",
        "name",
        "plural",
        "name_genitiv",
        "name_akuzativ",
    )
    list_display = ("display_name",)
    filter_horizontal = ("questions", "registration_agreements")
    exclude = ACTIVITY_TYPE_TEXT_FIELDS + ACTIVITY_TYPE_PRINT_SETUP_FIELDS
    prepopulated_fields = {"slug": ("plural",)}
    inlines = (ActivityTypeAttachmentInlineAdmin,)

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)

        # limit choices of registration agreements
        registration_agreements_choices = form.base_fields["registration_agreements"].widget.widget.choices
        registration_agreements_choices.queryset = registration_agreements_choices.queryset.exclude(
            id__in=request.leprikon_site.registration_agreements.values("id")
        )
        form.base_fields["registration_agreements"].choices = registration_agreements_choices

        return form

    @admin.display(description=_("activity type"))
    def display_name(self, obj: ActivityType) -> str:
        return obj.plural


@admin.register(ActivityTypeTexts)
class ActivityTypeTextsAdmin(ChildModelAdmin):
    parent_model = ActivityType
    fields = ACTIVITY_TYPE_TEXT_FIELDS


@admin.register(ActivityTypePrintSetups)
class ActivityTypePrintSetupsAdmin(ChildModelAdmin):
    parent_model = ActivityType
    fields = ACTIVITY_TYPE_PRINT_SETUP_FIELDS


@admin.register(ActivityGroup)
class ActivityGroupAdmin(SortableAdminMixin, admin.ModelAdmin):
    list_display = ("name", "color")
    list_editable = ("color",)
    filter_horizontal = ("activity_types",)


class ActivityAttachmentInlineAdmin(SortableInlineAdminMixin, admin.TabularInline):
    model = ActivityAttachment
    extra = 0


class ActivityTimeInlineAdmin(admin.TabularInline):
    model = ActivityTime
    extra = 0


class ActivityBaseAdmin(AdminExportMixin, BulkUpdateMixin, SendMessageAdminMixin, admin.ModelAdmin):
    registration_model = None
    actions = (
        SendMessageAdminMixin.actions + BulkUpdateMixin.actions + AdminExportMixin.actions + ("set_registration_dates",)
    )
    list_editable = ("note",)
    list_filter = (
        ("school_year", SchoolYearListFilter),
        "department",
        ("activity_type", ActivityTypeListFilter),
        ("groups", ActivityGroupListFilter),
        ("leaders", LeaderListFilter),
        "place",
    )
    inlines = (
        ActivityTimeInlineAdmin,
        ActivityAttachmentInlineAdmin,
    )
    filter_horizontal = ("age_groups", "target_groups", "groups", "leaders", "questions", "registration_agreements")
    search_fields = ("name", "description")
    save_as = True

    class Media:
        js = ["leprikon/js/Popup.js"]

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if not object_id and request.method == "POST" and len(request.POST) == 4:
            return HttpResponseRedirect(
                "{}?activity_type={}&registration_type={}".format(
                    request.path,
                    request.POST.get("activity_type", ""),
                    request.POST.get("registration_type", ""),
                )
            )
        else:
            return super().changeform_view(request, object_id, form_url, extra_context)

    def get_exclude(self, request, obj=None):
        if request.registration_type == Activity.PARTICIPANTS:
            exclude = ["target_groups"]
        elif request.registration_type == Activity.GROUPS:
            exclude = ["age_groups"]
        else:
            exclude = []
        if obj and obj.registrations.exists():
            exclude.append("registration_type")
        return exclude

    def get_fieldsets(self, request, obj: Activity = None):
        available_fields = self.get_fields(request, obj)
        used_fields = []
        fieldsets = []
        for label, all_fields in [
            (
                None,
                [
                    "activity_type",
                    "registration_type",
                    "name",
                    "description",
                    # course specific
                    "school_year_division",
                    "allow_period_selection",
                    # event specific
                    "start_date",
                    "end_date",
                    "start_time",
                    "end_time",
                    "due_from",
                    "due_date",
                    # orderable specific
                    "duration",
                    "due_from_days",
                    "due_date_days",
                    "min_due_date_days",
                    "age_groups",
                    "target_groups",
                    "reg_from",
                    "reg_to",
                    "public",
                ],
            ),
            (
                _("advanced settings"),
                [
                    "code",
                    "department",
                    "groups",
                    "place",
                    "leaders",
                    "photo",
                    "page",
                    "min_registrations_count",
                    "max_registrations_count",
                    "note",
                    "questions",
                    "registration_agreements",
                    "reg_print_setup",
                    "decision_print_setup",
                    "pr_print_setup",
                    "bill_print_setup",
                    "organization",
                ],
            ),
            (
                _("texts"),
                [
                    "text_registration_received",
                    "text_registration_approved",
                    "text_registration_refused",
                    "text_registration_payment_request",
                    "text_registration_refund_offer",
                    "text_registration_canceled",
                    "text_discount_granted",
                    "text_payment_received",
                    "text_payment_returned",
                ],
            ),
        ]:
            fields = [f for f in all_fields if f in available_fields]
            if fields:
                fieldsets.append((label, {"fields": fields}))
                used_fields += fields
        unused_fields = [f for f in available_fields if f not in used_fields]
        if unused_fields:
            fieldsets.append((None, {"fields": unused_fields}))
        return fieldsets

    def get_form(self, request, obj, **kwargs):
        # set school year
        if obj:
            request.school_year = obj.school_year

        # get activity type
        try:
            # first try request.POST (user may want to change activity type)
            request.activity_type = ActivityType.objects.get(id=int(request.POST.get("activity_type")))
        except (ActivityType.DoesNotExist, TypeError, ValueError):
            if obj:
                # use activity type from object
                request.activity_type = obj.activity_type
            else:
                # try to get activity type from request.GET
                try:
                    request.activity_type = ActivityType.objects.get(
                        id=int(request.GET.get("activity_type")),
                    )
                except (ActivityType.DoesNotExist, TypeError, ValueError):
                    request.activity_type = None

        # get registration type
        request.registration_type = request.POST.get("registration_type")
        if request.registration_type not in Activity.REGISTRATION_TYPES:
            if obj:
                # use registration type from object
                request.registration_type = obj.registration_type
            else:
                # try to get registration type from request.GET
                request.registration_type = request.GET.get("registration_type")
                if request.registration_type not in Activity.REGISTRATION_TYPES:
                    request.registration_type = None

        if request.activity_type and request.registration_type:
            kwargs["form"] = type(
                ActivityAdminForm.__name__,
                (ActivityAdminForm,),
                {
                    "school_year": request.school_year,
                    "activity_type": request.activity_type,
                    "registration_type": request.registration_type,
                },
            )
        else:
            kwargs["fields"] = ["activity_type", "registration_type"]
            request.hide_inlines = True

        form = super().get_form(request, obj, **kwargs)

        if request.activity_type and request.registration_type:
            for field_name in [
                "text_registration_received",
                "text_registration_approved",
                "text_registration_refused",
                "text_registration_payment_request",
                "text_registration_refund_offer",
                "text_registration_canceled",
                "text_discount_granted",
                "text_payment_received",
                "text_payment_returned",
            ]:
                form.base_fields[field_name].help_text = lazy_help_text_with_html_default(
                    form.base_fields[field_name].help_text,
                    getattr(request.activity_type, field_name) or DEFAULT_TEXTS[field_name],
                )

        return form

    def get_inline_instances(self, request, obj=None):
        return [] if hasattr(request, "hide_inlines") else super().get_inline_instances(request, obj)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        formfield = super().formfield_for_foreignkey(db_field, request, **kwargs)
        if db_field.name == "activity_type":
            limit_choices_to = {"model__exact": self.activity_type_model}
            formfield.limit_choices_to = limit_choices_to
        return formfield

    def save_form(self, request, form, change):
        obj = super().save_form(request, form, change)
        obj.school_year = request.school_year
        return obj

    @attributes(short_description=_("journals"))
    def journals_link(self, obj):
        return mark_safe(
            format_html(
                '<a href="{url}">{journals}</a>',
                url=reverse("admin:leprikon_journal_changelist")
                + f"?activity__activity_type__id__exact={obj.activity_type_id}&activity__id__exact={obj.id}",
                journals=_("journals"),
            )
            + format_html(
                '<a href="{url}" style="background-position: 0 0" title="{title}">' '<img src="{icon}" alt="+"/></a>',
                url=reverse("admin:leprikon_journal_add") + "?activity={}".format(obj.id),
                title=_("add journal"),
                icon=static("admin/img/icon-addlink.svg"),
            )
        )

    def get_message_recipients(self, request, queryset):
        return (
            get_user_model()
            .objects.filter(
                leprikon_registrations__canceled=None,
                leprikon_registrations__activity__in=queryset,
            )
            .distinct()
        )

    @attributes(short_description=_("registering"))
    def get_registering_link(self, obj):
        return mark_safe(
            format_html(
                '<a href="{url}">{activity_variants}</a>',
                url=reverse("admin:leprikon_activityvariant_changelist") + f"?activity__exact={obj.id}",
                activity_variants=_("price and registering variants"),
            )
            + format_html(
                '<a href="{url}" style="background-position: 0 0" title="{title}">' '<img src="{icon}" alt="+"/></a>',
                url=reverse("admin:leprikon_activityvariant_add") + "?activity={}".format(obj.id),
                title=_("add price and registering variant"),
                icon=static("admin/img/icon-addlink.svg"),
            )
        )

    @attributes(short_description=_("registrations"))
    def get_registrations_link(self, obj):
        icon = False
        approved_registrations_count = obj.approved_registrations.count()
        unapproved_registrations_count = obj.unapproved_registrations.count()

        if approved_registrations_count == 0:
            title = _("There are no approved registrations for this {}.").format(obj.activity_type.name_akuzativ)
        elif obj.min_registrations_count is not None and approved_registrations_count < obj.min_registrations_count:
            title = _("The number of approved registrations is lower than {}.").format(obj.min_registrations_count)
        elif obj.max_registrations_count is not None and approved_registrations_count > obj.max_registrations_count:
            title = _("The number of approved registrations is greater than {}.").format(obj.max_registrations_count)
        else:
            icon = True
            title = ""
        return mark_safe(
            format_html(
                '<a href="{url}" title="{title}">{icon} {approved}{unapproved}</a>',
                url=reverse(
                    "admin:{}_{}_changelist".format(
                        self.registration_model._meta.app_label,
                        self.registration_model._meta.model_name,
                    )
                )
                + "?activity__id__exact={}".format(obj.id),
                title=title,
                icon=_boolean_icon(icon),
                approved=approved_registrations_count,
                unapproved=" + {}".format(unapproved_registrations_count) if unapproved_registrations_count else "",
            )
            + format_html(
                '<a class="popup-link" href="{url}" style="background-position: 0 0" title="{title}">'
                '<img src="{icon}" alt="+"/></a>',
                url=reverse(
                    "admin:{}_{}_add".format(
                        self.registration_model._meta.app_label,
                        self.registration_model._meta.model_name,
                    )
                ),
                title=_("add registration"),
                icon=static("admin/img/icon-addlink.svg"),
            )
        )

    @attributes(short_description=_("approved registrations count"))
    def get_approved_registrations_count(self, obj):
        return obj.registrations.filter(canceled=None).exclude(approved=None).count()

    @attributes(short_description=_("unapproved registrations count"))
    def get_unapproved_registrations_count(self, obj):
        return obj.registrations.filter(canceled=None, approved=None).count()

    @attributes(short_description=_("photo"))
    def icon(self, obj):
        try:
            return mark_safe('<img src="{}" alt="{}"/>'.format(obj.photo.icons["48"], obj.photo.label))
        except (AttributeError, KeyError):
            return ""


class ChangeformRedirectMixin:
    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if object_id:
            obj = self.get_object(request, unquote(object_id))
            if obj:
                return HttpResponseRedirect(obj.get_edit_url())
        return super().changeform_view(request, object_id, form_url, extra_context)


@admin.register(Activity)
class ActivityAdmin(AdminExportMixin, SendMessageAdminMixin, ChangeformRedirectMixin, admin.ModelAdmin):
    """Hidden admin used for raw id fields"""

    actions = SendMessageAdminMixin.actions + AdminExportMixin.actions
    list_display = (
        "id",
        "code",
        "name",
        "activity_type",
        "get_groups_list",
        "get_leaders_list",
        "icon",
    )
    list_filter = (
        ("school_year", SchoolYearListFilter),
        "department",
        "activity_type__model",
        ("activity_type", ActivityTypeListFilter),
        "registration_type",
        ("groups", ActivityGroupListFilter),
        ("leaders", LeaderListFilter),
    )
    search_fields = ("name", "description")

    def get_model_perms(self, request):
        return {}

    def has_add_permission(self, request, obj=None):
        return False

    @attributes(short_description=_("photo"))
    def icon(self, obj):
        try:
            return mark_safe('<img src="{}" alt="{}"/>'.format(obj.photo.icons["48"], obj.photo.label))
        except (AttributeError, KeyError):
            return ""


@admin.register(ActivityVariant)
class ActivityVariantAdmin(AdminExportMixin, BulkUpdateMixin, SendMessageAdminMixin, admin.ModelAdmin):
    actions = SendMessageAdminMixin.actions + BulkUpdateMixin.actions + AdminExportMixin.actions
    bulk_update_exclude = ("activity", "order")
    list_display = (
        "activity",
        "name",
        "reg_from",
        "reg_to",
        "registration_price",
        "participant_price",
        "school_year_division",
        "get_registrations_link",
        "order",
    )
    list_editable = ("order",)
    list_filter = (
        ("activity__school_year", SchoolYearListFilter),
        "activity__department",
        "activity__activity_type__model",
        ("activity__activity_type", ActivityTypeListFilter),
        ("activity", ActivityListFilter),
    )
    raw_id_fields = ("activity",)
    search_fields = ("name", "description", "activity__name", "activity__description")
    filter_horizontal = ("age_groups", "target_groups", "required_resources", "required_resource_groups")

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if not object_id and request.method == "POST" and "activity" in request.POST:
            return HttpResponseRedirect(f"{request.path}?activity={request.POST['activity']}")
        else:
            return super().changeform_view(request, object_id, form_url, extra_context)

    def get_form(self, request: HttpRequest, obj: ActivityVariant | None, **kwargs):
        if obj:
            request.school_year = obj.activity.school_year
            request.activity = obj.activity
        else:
            # try to get activity from request.GET
            try:
                request.activity = Activity.objects.get(
                    school_year=request.school_year,
                    id=int(request.GET.get("activity")),
                )
            except (Activity.DoesNotExist, TypeError, ValueError):
                request.activity = None

        if request.activity:
            kwargs["form"] = type(ActivityVariantForm.__name__, (ActivityVariantForm,), {"activity": request.activity})
        else:
            kwargs["fields"] = ["activity"]

        return super().get_form(request, obj, **kwargs)

    def get_bulk_update_form(self, request: HttpRequest, fields: List[str]):
        form = super().get_bulk_update_form(request, fields)
        if "school_year_division" in form.base_fields:
            form = type(
                form.__name__,
                (form,),
                {"school_year_division": forms.ModelChoiceField(request.school_year.divisions.all())},
            )
        return form

    def get_exclude(self, request: HttpRequest, obj: ActivityVariant | None) -> Any:
        if not request.activity:
            return []
        exclude = ["activity", "order"]
        if request.activity:
            if request.activity.registration_type_participants:
                exclude.append("target_groups")
            if request.activity.registration_type_groups:
                exclude.append("age_groups")
            if request.activity.activity_type.model != ActivityModel.COURSE:
                exclude.append("school_year_division")
        return exclude

    def get_message_recipients(self, request, queryset):
        return (
            get_user_model()
            .objects.filter(
                leprikon_registrations__canceled=None,
                leprikon_registrations__activity_variant__in=queryset,
            )
            .distinct()
        )

    @attributes(short_description=_("registrations"))
    def get_registrations_link(self, obj):
        approved_registrations_count = obj.approved_registrations.count()
        unapproved_registrations_count = obj.unapproved_registrations.count()
        registration_model = {
            ActivityModel.COURSE: "courseregistration",
            ActivityModel.EVENT: "eventregistration",
            ActivityModel.ORDERABLE: "orderableregistration",
        }[obj.activity.activity_type.model]

        return mark_safe(
            format_html(
                '<a href="{url}">{approved}{unapproved}</a>',
                url=reverse(f"admin:leprikon_{registration_model}_changelist")
                + "?activity__id__exact={}&activity_variant__id__exact={}".format(obj.activity_id, obj.id),
                approved=approved_registrations_count,
                unapproved=" + {}".format(unapproved_registrations_count) if unapproved_registrations_count else "",
            )
            + format_html(
                '<a class="popup-link" href="{url}" style="background-position: 0 0" title="{title}">'
                '<img src="{icon}" alt="+"/></a>',
                url=reverse(f"admin:leprikon_{registration_model}_add") + f"?activity_variant={obj.id}",
                title=_("add registration"),
                icon=static("admin/img/icon-addlink.svg"),
            )
        )


class RegistrationParticipantInlineAdmin(admin.StackedInline):
    model = RegistrationParticipant
    extra = 0

    def get_min_num(self, request, obj=None, **kwargs):
        return request.activity_variant.min_participants_count

    def get_max_num(self, request, obj=None, **kwargs):
        return request.activity_variant.max_participants_count

    def get_formset(self, request, obj=None, **kwargs):
        questions = obj.all_questions if obj else request.activity.all_questions
        fields = dict(("q_" + q.slug, q.get_field()) for q in questions)
        fields["activity"] = request.activity
        fields["obj"] = obj
        kwargs["form"] = type(RegistrationParticipantAdminForm.__name__, (RegistrationParticipantAdminForm,), fields)
        return super().get_formset(request, obj, **kwargs)


class RegistrationGroupInlineAdmin(admin.StackedInline):
    model = RegistrationGroup
    extra = 1
    min_num = 1
    max_num = 1

    def get_formset(self, request, obj=None, **kwargs):
        questions = obj.all_questions if obj else request.activity.all_questions
        fields = dict(("q_" + q.slug, q.get_field()) for q in questions)
        fields["activity"] = request.activity
        fields["obj"] = obj
        kwargs["form"] = type(RegistrationGroupAdminForm.__name__, (RegistrationGroupAdminForm,), fields)
        return super().get_formset(request, obj, **kwargs)


class RegistrationGroupMemberInlineAdmin(admin.StackedInline):
    model = RegistrationGroupMember
    extra = 0

    def get_min_num(self, request, obj=None, **kwargs):
        return request.activity_variant.min_participants_count

    def get_max_num(self, request, obj=None, **kwargs):
        return request.activity_variant.max_participants_count


class RegistrationBillingInfoInlineAdmin(admin.TabularInline):
    model = RegistrationBillingInfo
    max_num = 1
    extra = 0


class RegistrationBaseAdmin(AdminExportMixin, SendMailAdminMixin, SendMessageAdminMixin, admin.ModelAdmin):
    actions = (
        SendMessageAdminMixin.actions
        + SendMailAdminMixin.actions
        + AdminExportMixin.actions
        + (
            "approve",
            "refuse",
            "request_payment",
            "offer_refund",
            "generate_refund_request",
            "export_invoices_xml",
            "cancel",
            "cancel_cancelation_request",
        )
    )
    form = RegistrationAdminForm
    inlines = (RegistrationBillingInfoInlineAdmin,)
    list_display = (
        "variable_symbol",
        "download_tag",
        "decision",
        "activity_name",
        "participants_list_html",
        "price",
        "discounts",
        "total_price",
        "received_payments",
        "returned_payments",
        "created_with_by",
        "approved_with_by",
        "payment_requested_with_by",
        "refund_offered_with_by",
        "cancelation_requested_with_by",
        "canceled_with_by",
        "note",
        "random_number",
    )
    list_editable = ("note",)
    list_export = (
        "id",
        "variable_symbol",
        "slug",
        "user",
        "activity",
        "activity_variant",
        "price",
        "discounts_export",
        "total_price_export",
        "received_payments_export",
        "note",
        "created",
        "created_by",
        "payment_requested",
        "payment_requested_by",
        "refund_offered",
        "refund_offered_by",
        "approved",
        "approved_by",
        "cancelation_requested",
        "cancelation_requested_by",
        "canceled",
        "canceled_by",
        "agreement_options_list",
        "group_members_list",
        "participants__gender",
        "participants__first_name",
        "participants__last_name",
        "participants__birth_num",
        "participants__birth_date",
        "participants__gender",
        "participants__age_group",
        "participants__street",
        "participants__city",
        "participants__postal_code",
        "participants__citizenship",
        "participants__phone",
        "participants__email",
        "participants__school",
        "participants__school_other",
        "participants__school_class",
        "participants__health",
        "participants__answers",
        "participants__has_parent1",
        "participants__parent1_first_name",
        "participants__parent1_last_name",
        "participants__parent1_street",
        "participants__parent1_city",
        "participants__parent1_postal_code",
        "participants__parent1_phone",
        "participants__parent1_email",
        "participants__has_parent2",
        "participants__parent2_first_name",
        "participants__parent2_last_name",
        "participants__parent2_street",
        "participants__parent2_city",
        "participants__parent2_postal_code",
        "participants__parent2_phone",
        "participants__parent2_email",
        "group__name",
        "group__first_name",
        "group__last_name",
        "group__street",
        "group__city",
        "group__postal_code",
        "group__phone",
        "group__email",
        "group__school__name",
        "group__school_other",
        "group__school_class",
        "billing_info__name",
        "billing_info__street",
        "billing_info__city",
        "billing_info__postal_code",
        "billing_info__company_num",
        "billing_info__vat_number",
        "billing_info__contact_person",
        "billing_info__phone",
        "billing_info__email",
        "billing_info__employee",
    )
    list_filter = (
        ("activity__school_year", SchoolYearListFilter),
        "activity__department",
        ("activity__activity_type", ActivityTypeListFilter),
        "activity__registration_type",
        "activity__organization",
        ApprovedListFilter,
        CanceledListFilter,
        "registration_link",
        ("billing_info", IsNullFieldListFilter),
        "activity__groups",
        ("activity", ActivityListFilter),
        ("activity__leaders", LeaderListFilter),
        "activity__place",
    )
    search_fields = (
        "variable_symbol",
        "participants__birth_num",
        "participants__first_name",
        "participants__last_name",
        "participants__parent1_first_name",
        "participants__parent1_last_name",
        "participants__parent2_first_name",
        "participants__parent2_last_name",
        "group_members__first_name",
        "group_members__last_name",
    )
    ordering = ("-created",)
    exclude = ["activity"]
    raw_id_fields = (
        "activity_variant",
        "calendar_event",
        "user",
    )

    class Media:
        css = {"all": ["leprikon/css/registrations.changelist.css"]}
        js = ["leprikon/js/Popup.js"]

    def has_delete_permission(self, request, obj=None):
        if obj and obj.approved is None and obj.received_payments.count() == 0 and obj.returned_payments.count() == 0:
            return super().has_delete_permission(request, obj)
        else:
            return False

    def lookup_allowed(self, lookup: str, value: str) -> bool:
        return lookup in (
            "participants__birth_date",
            "participants__birth_num",
            "participants__first_name",
            "participants__last_name",
        ) or super().lookup_allowed(lookup, value)

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    def get_changelist(self, request, **kwargs):
        ChangeListBase = super().get_changelist(request, **kwargs)

        class ChangeList(ChangeListBase):
            def get_ordering(self, request, queryset):
                # Show registrations with cancelation request on the top
                # if not showing canceled ones.
                ordering = super().get_ordering(request, queryset)
                return (
                    ordering
                    if request.GET.get("canceled") == "yes"
                    else [
                        IsNull("cancelation_requested"),
                        *ordering,
                    ]
                )

        return ChangeList

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .prefetch_related(
                "discounts",
                "received_payments",
                "returned_payments",
            )
            .select_related(
                "activity",
                "user",
            )
            .annotate(
                random_number=Random(),
                activity__organization__id=Coalesce(
                    F("activity__organization_id"),
                    F("activity__activity_type__organization_id"),
                ),
            )
        )

    @attributes(short_description=_("Approve selected registrations"))
    def approve(self, request, queryset):
        for registration in queryset.all():
            try:
                registration.approve(request.user)
            except ValidationError as e:
                self.message_user(request, e.message, messages.ERROR)
            else:
                self.message_user(
                    request,
                    _("The registration {r} has been approved and the user has been notified.").format(r=registration),
                )

    @attributes(short_description=_("Refuse selected registrations"))
    def refuse(self, request, queryset):
        for registration in queryset.all():
            try:
                registration.refuse(request.user)
            except ValidationError as e:
                self.message_user(request, e.message, messages.ERROR)
            else:
                self.message_user(
                    request,
                    _("The registration {r} has been refused and the user has been notified.").format(r=registration),
                )

    @attributes(short_description=_("Request payment for selected registrations"))
    def request_payment(self, request, queryset):
        for registration in queryset.all():
            registration.request_payment(request.user)
        self.message_user(request, _("Payment was requested for selected registrations."))

    @attributes(short_description=_("Offer refund for selected registrations"))
    def offer_refund(self, request, queryset):
        for registration in queryset.all():
            registration.offer_refund(request.user)
        self.message_user(request, _("Refund was offered for selected registrations."))

    @attributes(short_description=_("Generate refund requests for selected registrations"))
    def generate_refund_request(self, request, queryset):
        for registration in queryset.filter(refund_request__bank_account=None):
            registration.generate_refund_request(request.user)
        self.message_user(request, _("Refund requests were generated for selected registrations."))

    @attributes(short_description=_("Export selected registrations as invoices in XML"))
    def export_invoices_xml(self, request, queryset):
        response = SimpleTemplateResponse("leprikon/invoices.xml", {"registrations": queryset}, content_type="text/xml")
        response["Content-Disposition"] = 'attachment; filename="invoices.xml"'
        return response

    @attributes(short_description=_("Cancel selected registrations"))
    def cancel(self, request, queryset):
        for registration in queryset.all():
            try:
                registration.cancel(request.user)
            except ValidationError as e:
                self.message_user(request, e.message, messages.ERROR)
            else:
                self.message_user(
                    request,
                    _("The registration {r} has been canceled and the user has been notified.").format(r=registration),
                )

    @attributes(short_description=_("Cancel cancelation request for selected registrations"))
    def cancel_cancelation_request(self, request, queryset):
        queryset.update(cancelation_requested=None, cancelation_requested_by=None)
        self.message_user(request, _("The cancelation requests have been removed from selected registrations."))

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        formfield = super().formfield_for_foreignkey(db_field, request, **kwargs)
        if db_field.name == "activity_variant":
            limit_choices_to = {"activity__activity_type__model__exact": self.model.activity_type}
            formfield.limit_choices_to = limit_choices_to
            formfield.widget.rel.limit_choices_to = limit_choices_to
        return formfield

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if not object_id and request.method == "POST" and "user" not in request.POST:
            return HttpResponseRedirect(
                "{}?activity_variant={}".format(request.path, request.POST.get("activity_variant", ""))
            )
        else:
            return super().changeform_view(request, object_id, form_url, extra_context)

    def get_form(self, request, obj: Optional[Registration], **kwargs):
        try:
            # first try request.POST (user may want to change the activity variant)
            request.activity_variant = ActivityVariant.objects.get(id=int(request.POST.get("activity_variant")))
        except (ActivityVariant.DoesNotExist, TypeError, ValueError):
            if obj:
                # use activity from object
                request.activity_variant = obj.activity_variant
            else:
                # try to get activity from request.GET
                try:
                    request.activity_variant = ActivityVariant.objects.get(id=int(request.GET.get("activity_variant")))
                except (Activity.DoesNotExist, TypeError, ValueError):
                    request.activity_variant = None

        request.activity = request.activity_variant and request.activity_variant.activity

        if request.activity_variant:
            kwargs["form"] = type(
                self.form.__name__,
                (self.form,),
                {"activity": request.activity, "activity_variant": request.activity_variant},
            )
        else:
            kwargs["form"] = forms.ModelForm
            kwargs["fields"] = ["activity_variant"]
        return super().get_form(request, obj, **kwargs)

    def get_inline_instances(self, request, obj=None):
        if request.activity:
            if request.activity.registration_type_participants:
                inlines = (RegistrationParticipantInlineAdmin,)
            elif request.activity.registration_type_groups:
                inlines = (RegistrationGroupInlineAdmin, RegistrationGroupMemberInlineAdmin)
            return [inline(self.model, self.admin_site) for inline in inlines] + super().get_inline_instances(
                request, obj
            )
        else:
            return []

    def save_model(self, request, obj, form, change):
        obj.activity = obj.activity_variant.activity
        if not change:
            if obj.activity.registration_type_participants:
                participants_count = int(form.data["participants-TOTAL_FORMS"])
            elif obj.activity.registration_type_groups:
                participants_count = int(form.data["group_members-TOTAL_FORMS"])
            obj.price = obj.activity_variant.get_price(participants_count)
            obj.created_by = request.user
        super().save_model(request, obj, form, change)

    def save_related(self, request, form, formsets, change):
        super().save_related(request, form, formsets, change)
        if not change:
            form.instance.questions.set(form.instance.activity.all_questions)
            form.instance.agreements.set(form.instance.activity.all_registration_agreements)
            form.instance.generate_variable_symbol_and_slug()
            form.instance.send_mail()

    def get_css(self, obj):
        classes = []
        if obj.cancelation_requested and not obj.canceled:
            classes.append("reg-cancel-request")
        if obj.approved:
            classes.append("reg-approved")
        else:
            classes.append("reg-new")
            if obj.activity.full:
                classes.append("reg-full")
        if obj.canceled:
            classes.append("reg-canceled")
        else:
            classes.append("reg-active")
        return " ".join(classes)

    legend = (
        ("reg-cancel-request", _("cancelation requested")),
        ("reg-new", _("new registration")),
        ("reg-new reg-full", _("new registration, capacity full")),
        ("reg-approved", _("approved registration")),
    )

    @attributes(short_description=_("activity"))
    def activity_name(self, obj):
        return obj.activity.name

    @attributes(short_description=_("decision"))
    def decision(self, obj):
        if obj.approved or obj.canceled:
            return mark_safe(
                '<a href="{}">{}</a>'.format(
                    reverse(
                        "admin:{}_{}_event_pdf".format(self.model._meta.app_label, self.model._meta.model_name),
                        args=(obj.id, "decision"),
                    ),
                    _("admission decision") if obj.approved else _("refusal decision"),
                )
            )

    created_with_by = datetime_with_by("created", _("time of registration"))

    approved_with_by = datetime_with_by("approved", _("time of approval"))

    payment_requested_with_by = datetime_with_by("payment_requested", _("payment request time"))

    refund_offered_with_by = datetime_with_by("refund_offered", _("refund offer time"))

    cancelation_requested_with_by = datetime_with_by("cancelation_requested", _("time of cancellation request"))

    canceled_with_by = datetime_with_by("canceled", _("time of cancellation"))

    def get_message_recipients(self, request, queryset):
        return get_user_model().objects.filter(leprikon_registrations__in=queryset).distinct()

    @attributes(admin_order_field="random_number", short_description=_("random number"))
    def random_number(self, obj):
        return int(obj.random_number * 1000000000000)

    @attributes(short_description=_("received payments"))
    def received_payments(self, obj: Registration):
        return mark_safe(
            format_html(
                '<a style="color: {color}" href="{href_list}" title="{title}"><b>{amount}</b></a>'
                ' &nbsp; <a class="popup-link" href="{href_add}" style="background-position: 0 0" title="{title_add}">'
                '<img src="{icon_add}" alt="+"/></a>',
                amount=currency(obj.payment_status.received),
                color=obj.payment_status.color,
                href_add=reverse("admin:leprikon_receivedpayment_add") + f"?target_registration={obj.id}",
                href_list=reverse("admin:leprikon_receivedpayment_changelist") + f"?target_registration={obj.id}",
                icon_add=static("admin/img/icon-addlink.svg"),
                title=obj.payment_status.title,
                title_add=_("add received payment"),
            )
        )

    @attributes(short_description=_("returned payments"))
    def returned_payments(self, obj: Registration):
        return mark_safe(
            format_html(
                '<a href="{href_list}"><b>{amount}</b></a>'
                ' &nbsp; <a class="popup-link" href="{href_add}" style="background-position: 0 0" title="{title_add}">'
                '<img src="{icon_add}" alt="+"/></a>',
                amount=currency(obj.payment_status.returned),
                href_add=reverse("admin:leprikon_returnedpayment_add") + f"?source_registration={obj.id}",
                href_list=reverse("admin:leprikon_returnedpayment_changelist") + f"?source_registration={obj.id}",
                icon_add=static("admin/img/icon-addlink.svg"),
                title_add=_("add returned payment"),
            )
        )

    @attributes(short_description=_("total price"))
    def total_price(self, obj: Registration):
        return currency(obj.payment_status.receivable)


@admin.register(Registration)
class RegistrationAdmin(AdminExportMixin, SendMessageAdminMixin, ChangeformRedirectMixin, admin.ModelAdmin):
    """Hidden admin used for raw id fields"""

    actions = SendMessageAdminMixin.actions + AdminExportMixin.actions
    list_display = (
        "id",
        "variable_symbol",
        "activity",
        "participants_list_html",
        "group",
        "created",
        "canceled",
    )
    list_filter = (
        ("activity__school_year", SchoolYearListFilter),
        "activity__department",
        ("activity__activity_type", ActivityTypeListFilter),
        ApprovedListFilter,
        CanceledListFilter,
        ("activity", ActivityListFilter),
        ("activity__leaders", LeaderListFilter),
    )
    ordering = ("-created",)
    search_fields = (
        "variable_symbol",
        "participants__birth_num",
        "participants__first_name",
        "participants__last_name",
        "participants__parent1_first_name",
        "participants__parent1_last_name",
        "participants__parent2_first_name",
        "participants__parent2_last_name",
        "group_members__first_name",
        "group_members__last_name",
    )

    def lookup_allowed(self, lookup: str, value: str) -> bool:
        return lookup in (
            "participants__birth_date",
            "participants__birth_num",
            "participants__first_name",
            "participants__last_name",
        ) or super().lookup_allowed(lookup, value)

    def get_model_perms(self, request):
        return {}

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_actions(self, request):
        return {}


class ActivityDiscountBaseAdmin(TransactionAdminMixin, AdminExportMixin, admin.ModelAdmin):
    list_filter = (
        ("registration__activity__school_year", SchoolYearListFilter),
        "registration__activity__department",
        ("registration__activity__activity_type", ActivityTypeListFilter),
        ("registration__activity", ActivityListFilter),
        ("registration__activity__leaders", LeaderListFilter),
    )
    search_fields = (
        "registration__variable_symbol",
        "registration__activity__name",
        "registration__participants__first_name",
        "registration__participants__last_name",
        "registration__participants__birth_num",
    )
    raw_id_fields = ("registration",)
    closed_fields = ("accounted", "registration", "amount")

    @attributes(short_description=_("activity"))
    def activity(self, obj):
        return obj.registration.activity

    list_export = (
        "id",
        "accounted",
        "accounted_by",
        "transaction_type_label",
        "amount",
        "organization",
        "donor__username",
        "donor",
        "last_updated",
        "last_updated_by",
        "mail_sent",
        "note",
    )


class PaymentAdminMixin:
    list_display = (
        "id",
        "accounted_with_by",
        "download_tag",
        "transaction_type",
        "amount_html",
        "registration",
        "last_updated_with_by",
        "mail_sent",
        "note",
    )
    list_export = (
        "id",
        "accounted",
        "accounted_by",
        "transaction_type_label",
        "amount",
        "registration",
        "variable_symbol",
        "last_updated",
        "last_updated_by",
        "note",
    )

    @attributes(short_description=_("variable symbol"))
    def variable_symbol(self, obj):
        return obj.registration.variable_symbol


@admin.register(ReceivedPayment)
class ReceivedPaymentAdmin(PaymentAdminMixin, TransactionBaseAdmin):
    exclude = ("donor", "organization")
    list_filter = (
        ("target_registration__activity__school_year", SchoolYearListFilter),
        ("transaction_type", TransactionTypeListFilter),
        "target_registration__activity__department",
        ("target_registration__activity__activity_type", ActivityTypeListFilter),
        ("target_registration__activity", ActivityListFilter),
        ("target_registration__activity__leaders", LeaderListFilter),
    )
    search_fields = (
        "target_registration__activity__name",
        "target_registration__participants__first_name",
        "target_registration__participants__last_name",
        "target_registration__participants__birth_num",
    )

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        form.base_fields["target_registration"].required = True
        return form


@admin.register(ReturnedPayment)
class ReturnedPaymentAdmin(PaymentAdminMixin, TransactionBaseAdmin):
    exclude = ("pays_payment",)
    list_filter = (
        ("source_registration__activity__school_year", SchoolYearListFilter),
        ("transaction_type", TransactionTypeListFilter),
        "source_registration__activity__department",
        ("source_registration__activity__activity_type", ActivityTypeListFilter),
        ("source_registration__activity", ActivityListFilter),
        ("source_registration__activity__leaders", LeaderListFilter),
    )
    search_fields = (
        "source_registration__activity__name",
        "source_registration__participants__first_name",
        "source_registration__participants__last_name",
        "source_registration__participants__birth_num",
    )

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        form.base_fields["source_registration"].required = True
        return form

    @attributes(admin_order_field="amount", short_description=_("amount"))
    def amount_html(self, obj):
        return mark_safe(
            format_html(
                '<b style="color: {color}">{amount}</b>',
                color=amount_color(-obj.amount),
                amount=currency(obj.amount),
            )
        )



from datetime import time

from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


class StartEndMixin(object):
    def validate(self):
        # (time(0) represents the end of the day)
        for start_field, end_field in (
            ("start", "end"),
            ("start_time", "end_time"),
            ("start_date", "end_date"),
        ):
            start = getattr(self, start_field, None)
            end = getattr(self, end_field, None)
            if start and end and start > end and end != time(0):
                raise ValidationError(
                    {
                        start_field: [_("Start must be before end")],
                        end_field: [_("End must be later than start")],
                    }
                )

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude)
        self.validate()

    def save(self, *args, **kwargs):
        self.validate()
        super().save(*args, **kwargs)



# Generated by Django 1.11.13 on 2018-09-04 18:26

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
from django.utils.timezone import utc
import djangocms_text_ckeditor.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('leprikon', '0020_agreements'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserAgreement',
            fields=[
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='agreement', serialize=False, to=settings.AUTH_USER_MODEL, verbose_name='user')),
                ('granted', models.DateTimeField(auto_now=True, verbose_name='time of agreement')),
            ],
        ),
        migrations.RenameField(
            model_name='leprikonsite',
            old_name='agreement',
            new_name='registration_agreement',
        ),
        migrations.AddField(
            model_name='leprikonsite',
            name='user_agreement',
            field=djangocms_text_ckeditor.fields.HTMLField(blank=True, default='', verbose_name='user agreement'),
        ),
        migrations.AddField(
            model_name='leprikonsite',
            name='user_agreement_changed',
            field=models.DateTimeField(blank=True, null=True, verbose_name='last time user agreement changed'),
            preserve_default=False,
        ),
    ]



from collections import namedtuple
from datetime import date, datetime, timedelta

from cms.models import CMSPlugin
from django.db import models
from django.utils.formats import date_format, time_format
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from ..utils import attributes
from .activities import (
    Activity,
    ActivityDiscount,
    ActivityGroup,
    ActivityModel,
    ActivityType,
    ActivityVariant,
    Registration,
)
from .agegroup import AgeGroup
from .department import Department
from .roles import Leader
from .schoolyear import SchoolYear, SchoolYearDivision
from .targetgroup import TargetGroup
from .utils import PaymentStatus, change_year, copy_related_objects


class Orderable(Activity):
    duration = models.DurationField(_("duration"), help_text=_("Format: HH:MM:SS"))
    preparation_time = models.DurationField(
        _("preparation time"),
        default=timedelta(0),
        help_text=_("Time to prepare before the event. (HH:MM:SS)"),
    )
    recovery_time = models.DurationField(
        _("recovery time"),
        default=timedelta(0),
        help_text=_("Time to recover after the event. (HH:MM:SS)"),
    )
    due_from_days = models.IntegerField(
        _("number of days to send the payment request before event date"),
        blank=True,
        null=True,
        help_text=_(
            "If set, payment request will be sent this number of days before event date. "
            "If not set, payment request will be sent when registration is approved."
        ),
    )
    due_date_days = models.IntegerField(_("number of days to due date before event date"), default=0)
    min_due_date_days = models.PositiveIntegerField(_("minimal number of days to due date"), default=3)

    class Meta:
        app_label = "leprikon"
        ordering = ("code", "name")
        verbose_name = _("orderable event")
        verbose_name_plural = _("orderable events")

    def copy_to_school_year(old, school_year: SchoolYear):
        new = Orderable.objects.get(id=old.id)
        new.id, new.pk = None, None
        new.school_year = school_year
        new.public = False
        new.note = ""
        year_delta = school_year.year - old.school_year.year
        new.save()
        new.groups.set(old.groups.all())
        new.age_groups.set(old.age_groups.all())
        new.target_groups.set(old.target_groups.all())
        for leader in old.all_leaders:
            school_year.leaders.add(leader)
        new.leaders.set(old.all_leaders)
        new.questions.set(old.questions.all())
        for old_variant in old.all_variants:
            new_variant = ActivityVariant.objects.get(id=old_variant.id)
            new_variant.id, new_variant.pk = None, None
            new_variant.activity = new
            if old_variant.school_year_division:
                new_variant.school_year_division = SchoolYearDivision.objects.filter(
                    school_year=school_year,
                    name=old_variant.school_year_division.name,
                ).first() or old_variant.school_year_division.copy_to_school_year(school_year)
            new_variant.reg_from = change_year(new_variant.reg_from, year_delta)
            new_variant.reg_to = change_year(new_variant.reg_to, year_delta)
            new_variant.save()
            new_variant.age_groups.set(old_variant.age_groups.all())
            new_variant.target_groups.set(old_variant.target_groups.all())
        copy_related_objects(
            new,
            attachments=old.attachments,
            times=old.times,
        )
        return new


class OrderableRegistration(Registration):
    activity_type = ActivityModel.ORDERABLE

    class Meta:
        app_label = "leprikon"
        verbose_name = _("orderable event registration")
        verbose_name_plural = _("orderable event registrations")

    def get_payment_status(self, d=None):
        payment_status = PaymentStatus(
            price=self.price,
            discount=self.get_discounted(d),
            explanation=",\n".join(
                discount.explanation.strip()
                for discount in self.all_discounts
                if (d is None or discount.accounted.date() <= d) and discount.explanation.strip()
            ),
            received=self.get_received(d),
            returned=self.get_returned(d),
            current_date=d or date.today(),
            due_from=self.payment_requested
            and (
                self.payment_requested.date()
                if self.activity.orderable.due_from_days is None
                else max(
                    self.start_date - timedelta(days=self.activity.orderable.due_from_days),
                    self.payment_requested.date(),
                )
            ),
            due_date=self.payment_requested
            and max(
                self.start_date - timedelta(days=self.activity.orderable.due_date_days),
                self.payment_requested.date() + timedelta(days=self.activity.orderable.min_due_date_days),
            ),
        )
        if d is None and self.cached_balance != payment_status.balance:
            self.cached_balance = payment_status.balance
            self.save(update_fields=["cached_balance"])
        return payment_status

    @attributes(admin_order_field="start_date", short_description=_("event date"))
    def event_date(self):
        return "{start}{separator}{end}".format(
            start=(
                date_format(datetime.combine(self.start_date, self.start_time), "SHORT_DATETIME_FORMAT")
                if self.start_time
                else date_format(self.start_date, "SHORT_DATE_FORMAT")
            ),
            separator=" - " if self.start_date != self.end_date or self.end_time is not None else "",
            end=(
                (time_format(self.end_time, "TIME_FORMAT") if self.end_time else "")
                if self.start_date == self.end_date
                else (
                    date_format(datetime.combine(self.end_date, self.end_time), "SHORT_DATETIME_FORMAT")
                    if self.end_time
                    else date_format(self.end_date, "SHORT_DATE_FORMAT")
                )
            ),
        )


class OrderableDiscount(ActivityDiscount):
    registration = models.ForeignKey(
        OrderableRegistration, on_delete=models.CASCADE, related_name="discounts", verbose_name=_("registration")
    )

    class Meta:
        app_label = "leprikon"
        verbose_name = _("orderable event discount")
        verbose_name_plural = _("orderable event discounts")
        ordering = ("accounted",)


class OrderablePlugin(CMSPlugin):
    event = models.ForeignKey(Orderable, on_delete=models.CASCADE, related_name="+", verbose_name=_("event"))
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_ORDERABLE_TEMPLATES,
        default=settings.LEPRIKON_ORDERABLE_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"


class OrderableListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    departments = models.ManyToManyField(
        Department,
        blank=True,
        related_name="+",
        verbose_name=_("departments"),
        help_text=_("Keep empty to skip searching by departments."),
    )
    event_types = models.ManyToManyField(
        ActivityType,
        blank=True,
        limit_choices_to={"activity_type": ActivityModel.ORDERABLE},
        related_name="+",
        verbose_name=_("event types"),
        help_text=_("Keep empty to skip searching by event types."),
    )
    age_groups = models.ManyToManyField(
        AgeGroup,
        blank=True,
        related_name="+",
        verbose_name=_("age groups"),
        help_text=_("Keep empty to skip searching by age groups."),
    )
    target_groups = models.ManyToManyField(
        TargetGroup,
        blank=True,
        related_name="+",
        verbose_name=_("target groups"),
        help_text=_("Keep empty to skip searching by target groups."),
    )
    groups = models.ManyToManyField(
        ActivityGroup,
        blank=True,
        related_name="+",
        verbose_name=_("event groups"),
        help_text=_("Keep empty to skip searching by groups."),
    )
    leaders = models.ManyToManyField(
        Leader,
        verbose_name=_("leaders"),
        blank=True,
        related_name="+",
        help_text=_("Keep empty to skip searching by leaders."),
    )
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_ORDERABLELIST_TEMPLATES,
        default=settings.LEPRIKON_ORDERABLELIST_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.departments.set(oldinstance.departments.all())
        self.event_types.set(oldinstance.event_types.all())
        self.groups.set(oldinstance.groups.all())
        self.age_groups.set(oldinstance.age_groups.all())
        self.target_groups.set(oldinstance.age_groups.all())
        self.leaders.set(oldinstance.leaders.all())

    @cached_property
    def all_departments(self):
        return list(self.departments.all())

    @cached_property
    def all_event_types(self):
        return list(self.event_types.all())

    @cached_property
    def all_groups(self):
        return list(self.groups.all())

    @cached_property
    def all_age_groups(self):
        return list(self.age_groups.all())

    @cached_property
    def all_target_groups(self):
        return list(self.target_groups.all())

    @cached_property
    def all_leaders(self):
        return list(self.leaders.all())

    Group = namedtuple("Group", ("group", "objects"))

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )
        events = Orderable.objects.filter(school_year=school_year, public=True).distinct()

        if self.all_departments:
            events = events.filter(department__in=self.all_departments)
        if self.all_event_types:
            events = events.filter(activity_type__in=self.all_event_types)
        if self.all_age_groups:
            events = events.filter(age_groups__in=self.all_age_groups)
        if self.all_target_groups:
            events = events.filter(target_groups__in=self.all_target_groups)
        if self.all_leaders:
            events = events.filter(leaders__in=self.all_leaders)
        if self.all_groups:
            events = events.filter(groups__in=self.all_groups)
            groups = self.all_groups
        elif self.all_event_types:
            groups = ActivityGroup.objects.filter(activity_types__in=self.all_event_types)
        else:
            groups = ActivityGroup.objects.all()

        context.update(
            {
                "school_year": school_year,
                "events": events,
                "groups": (self.Group(group=group, objects=events.filter(groups=group)) for group in groups),
            }
        )
        return context


class FilteredOrderableListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    event_types = models.ManyToManyField(
        ActivityType,
        limit_choices_to={"activity_type": ActivityModel.ORDERABLE},
        related_name="+",
        verbose_name=_("event types"),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.event_types = oldinstance.event_types.all()

    @cached_property
    def all_event_types(self):
        return list(self.event_types.all())

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )

        from ..forms.activities import ActivityFilterForm

        form = ActivityFilterForm(
            activity_type_model=ActivityModel.ORDERABLE,
            activity_types=self.all_event_types,
            school_year=school_year,
            is_staff=context["request"].user.is_staff,
            data=context["request"].GET,
        )
        context.update(
            {
                "school_year": school_year,
                "form": form,
                "events": form.get_queryset(),
            }
        )
        return context



from django.utils.translation import gettext_lazy as _

PRICE_DECIMAL_PLACES = 2
PRICE_MAX_DIGITS = 10

LEPRIKON_SHOW_ACTIVITY_CODE = False

LEPRIKON_COUNTRY = "CZ"

LEPRIKON_CONTACT_TYPES = (
    ("email", _("email")),
    ("phone", _("phone")),
    ("address", _("address")),
    ("url", _("url")),
)

LEPRIKON_COURSE_TEMPLATES = (("default", _("Default")),)

LEPRIKON_EVENT_TEMPLATES = (("default", _("Default")),)

LEPRIKON_ORDERABLE_TEMPLATES = (("default", _("Default")),)

LEPRIKON_LEADER_TEMPLATES = (("default", _("Default")),)

LEPRIKON_COURSELIST_TEMPLATES = (
    ("default", _("Default")),
    ("grouped", _("Grouped by course groups")),
)

LEPRIKON_EVENTLIST_TEMPLATES = (
    ("default", _("Default")),
    ("grouped", _("Grouped by event groups")),
)

LEPRIKON_ORDERABLELIST_TEMPLATES = (
    ("default", _("Default")),
    ("grouped", _("Grouped by event groups")),
)

LEPRIKON_LEADERLIST_TEMPLATES = (("default", _("Default")),)

LEPRIKON_COLOR_POSITIVE = LEPRIKON_COLOR_PAID = "#0a0"
LEPRIKON_COLOR_NEGATIVE = LEPRIKON_COLOR_NOTPAID = "#e60"
LEPRIKON_COLOR_ZERO = LEPRIKON_COLOR_OVERPAID = "#00a"
LEPRIKON_COLOR_OVERDUE = "#e00"

LEPRIKON_QUESTION_FIELDS = {
    "boolean": {
        "name": _("boolean field"),
        "class": "django.forms.BooleanField",
    },
    "char": {
        "name": _("char field"),
        "class": "django.forms.CharField",
    },
    "text": {
        "name": _("text field"),
        "class": "leprikon.forms.fields.TextField",
    },
    "choice": {
        "name": _("choice field"),
        "class": "django.forms.ChoiceField",
    },
    "required_optional": {
        "name": _("required field with an empty option"),
        "class": "leprikon.forms.fields.RequiredOptionalField",
    },
    "date": {
        "name": _("date field"),
        "class": "django.forms.DateField",
    },
}

LEPRIKON_PARAM_BACK = "zpet"

LEPRIKON_URL = None

LEPRIKON_URL_SUMMARY = "prehled"
LEPRIKON_URL_LEADER = "pro-vedouci"
LEPRIKON_URL_CREATE_ACCOUNT = "novy-uzivatel"
LEPRIKON_URL_COURSES = "kurzy"
LEPRIKON_URL_EVENTS = "akce"
LEPRIKON_URL_ORDERABLES = "na-objednani"
LEPRIKON_URL_LEADERS = "vedouci"
LEPRIKON_URL_MESSAGES = "zpravy"
LEPRIKON_URL_TIMESHEETS = "vykazy"
LEPRIKON_URL_SUBMIT = "odevzdat"
LEPRIKON_URL_ENTRY = "zaznam"
LEPRIKON_URL_PARENT = "zastupce"
LEPRIKON_URL_PARTICIPANT = "ucastnik"
LEPRIKON_URL_PARTICIPANTS = "ucastnici"
LEPRIKON_URL_BILLING_INFO = "fakturacni-udaje"
LEPRIKON_URL_USER = "uzivatel"
LEPRIKON_URL_AGREEMENT = "souhlas"
LEPRIKON_URL_PASSWORD = "heslo"
LEPRIKON_URL_PASSWORD_RESET = "reset-hesla"
LEPRIKON_URL_LOGIN = "prihlasit"
LEPRIKON_URL_LOGOUT = "odhlasit"
LEPRIKON_URL_SCHOOL_YEAR = "skolni-rok"
LEPRIKON_URL_SUPPORT = "pomoc"
LEPRIKON_URL_EDIT = "upravit"
LEPRIKON_URL_JOURNAL = "denik"
LEPRIKON_URL_JOURNALS = "deniky"
LEPRIKON_URL_JOURNALENTRY = "zaznam-deniku"
LEPRIKON_URL_MINE = "moje"
LEPRIKON_URL_ALTERNATING = "zastup"
LEPRIKON_URL_REPORTS = "reporty"
LEPRIKON_URL_PAYMENT_REQUEST = "vyzva-k-platbe"
LEPRIKON_URL_PAYMENTS = "platby"
LEPRIKON_URL_RECEIVED_PAYMENTS = "prijate-platby"
LEPRIKON_URL_RETURNED_PAYMENTS = "vracene-platby"
LEPRIKON_URL_REFUND_REQUESTS = "vraceni-preplatku"
LEPRIKON_URL_PAYMENT_TRANSFER = "prevedeni-platby"
LEPRIKON_URL_DONATIONS = "dary"
LEPRIKON_URL_PAYMENTS_STATUS = "stav-plateb"
LEPRIKON_URL_DEBTORS = "dluznici"
LEPRIKON_URL_ADD = "pridat"
LEPRIKON_URL_DELETE = "smazat"
LEPRIKON_URL_CANCEL = "zrusit"
LEPRIKON_URL_REGISTRATION = "prihlaska"
LEPRIKON_URL_REGISTRATIONS = "prihlasky"
LEPRIKON_URL_REGISTRATION_LINK = "prihlasovaci-odkaz"
LEPRIKON_URL_TERMS_CONDITIONS = "obchodni-podminky"
LEPRIKON_URL_STATS = "statistiky"
LEPRIKON_URL_EMAIL = "email"

LEPRIKON_MENU_ADD_LOGOUT = True

LEPRIKON_SITE_CACHE_MAX_AGE = 5

# expression to create variable symbol (activity.code + last two digits of year + last four digits of id)
LEPRIKON_VARIABLE_SYMBOL_EXPRESSION = (
    "reg.activity.code * 1000000 + (reg.created.year % 100) * 10000 + (reg.id % 10000)"
)

SESSION_STAFF_COOKIE_AGE = 60 * 60 * 10



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-12-01 19:04
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import leprikon.models.fields
import localflavor.generic.models


def migrate_organization(apps, schema_editor):
    Organization = apps.get_model("leprikon", "Organization")
    LeprikonSite = apps.get_model("leprikon", "LeprikonSite")

    for leprikon_site in LeprikonSite.objects.all():
        leprikon_site.organization = Organization.objects.create(
            name=leprikon_site.company_name or leprikon_site.name,
            street=leprikon_site.street,
            city=leprikon_site.city,
            postal_code=leprikon_site.postal_code,
            email=leprikon_site.email,
            phone=leprikon_site.phone,
            company_num=leprikon_site.company_num,
            vat_number=leprikon_site.vat_number,
            iban=leprikon_site.iban,
            bic=leprikon_site.bic,
        )
        leprikon_site.save()


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0037_message_sender"),
    ]

    operations = [
        migrations.CreateModel(
            name="Organization",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=150, verbose_name="organization name")),
                ("street", models.CharField(blank=True, max_length=150, null=True, verbose_name="street")),
                ("city", models.CharField(blank=True, max_length=150, null=True, verbose_name="city")),
                (
                    "postal_code",
                    leprikon.models.fields.PostalCodeField(blank=True, null=True, verbose_name="postal code"),
                ),
                (
                    "email",
                    leprikon.models.fields.EmailField(
                        blank=True, max_length=254, null=True, verbose_name="email address"
                    ),
                ),
                ("phone", models.CharField(blank=True, max_length=30, null=True, verbose_name="phone")),
                ("company_num", models.CharField(blank=True, max_length=8, null=True, verbose_name="company number")),
                ("vat_number", models.CharField(blank=True, max_length=10, null=True, verbose_name="VAT number")),
                (
                    "iban",
                    localflavor.generic.models.IBANField(
                        blank=True,
                        include_countries=None,
                        max_length=34,
                        null=True,
                        use_nordea_extensions=False,
                        verbose_name="IBAN",
                    ),
                ),
                (
                    "bic",
                    localflavor.generic.models.BICField(
                        blank=True, max_length=11, null=True, verbose_name="BIC (SWIFT)"
                    ),
                ),
            ],
            options={
                "verbose_name": "organization",
                "verbose_name_plural": "organizations",
            },
        ),
        migrations.AddField(
            model_name="leprikonsite",
            name="organization",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.Organization",
                verbose_name="default organization",
            ),
        ),
        migrations.AddField(
            model_name="subject",
            name="organization",
            field=models.ForeignKey(
                blank=True,
                help_text="Only use to set value specific for this subject.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.Organization",
                verbose_name="organization",
            ),
        ),
        migrations.AddField(
            model_name="subjecttype",
            name="organization",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.Organization",
                verbose_name="organization",
            ),
        ),
        migrations.RunPython(migrate_organization),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="bic",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="city",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="company_name",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="company_num",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="email",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="iban",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="phone",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="postal_code",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="street",
        ),
        migrations.RemoveField(
            model_name="leprikonsite",
            name="vat_number",
        ),
        migrations.AlterField(
            model_name="leprikonsite",
            name="bill_print_setup",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.PrintSetup",
                verbose_name="default bill print setup",
            ),
        ),
        migrations.AlterField(
            model_name="leprikonsite",
            name="reg_print_setup",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.PrintSetup",
                verbose_name="default registration print setup",
            ),
        ),
        migrations.AlterField(
            model_name="leprikonsite",
            name="registration_agreements",
            field=models.ManyToManyField(
                blank=True,
                help_text="Add legal agreements for the registration form.",
                related_name="+",
                to="leprikon.Agreement",
                verbose_name="default registration agreements",
            ),
        ),
        migrations.AlterField(
            model_name="subject",
            name="bill_print_setup",
            field=models.ForeignKey(
                blank=True,
                help_text="Only use to set value specific for this subject.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.PrintSetup",
                verbose_name="payment print setup",
            ),
        ),
        migrations.AlterField(
            model_name="subject",
            name="reg_print_setup",
            field=models.ForeignKey(
                blank=True,
                help_text="Only use to set value specific for this subject.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="leprikon.PrintSetup",
                verbose_name="registration print setup",
            ),
        ),
    ]



from django import forms
from django.contrib.admin import widgets
from django.contrib.auth import get_user_model
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from ..models.messages import Message, MessageRecipient
from .form import FormMixin


class MessageFilterForm(FormMixin, forms.Form):
    q = forms.CharField(label=_("Search term"), required=False)

    def __init__(self, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qs = user.leprikon_messages.all()

    def get_queryset(self):
        if not self.is_valid():
            return self.qs
        qs = self.qs
        for word in self.cleaned_data["q"].split():
            qs = qs.filter(Q(message__subject__icontains=word) | Q(message__text__icontains=word))
        return qs.distinct()


class MessageAdminForm(forms.ModelForm):
    recipients = forms.MultipleChoiceField(
        choices=lambda: [
            (u.id, "{} ({})".format(u.get_username(), u.get_full_name())) for u in get_user_model().objects.all()
        ],
        widget=widgets.FilteredSelectMultiple(
            verbose_name=_("Recipients"),
            is_stacked=False,
        ),
        label=_("Recipients"),
        required=False,
    )

    class Meta:
        model = Message
        fields = ["sender", "subject", "text"]

    def __init__(self, *args, **kwargs):
        instance = kwargs.get("instance", None)
        if instance:
            kwargs["initial"] = kwargs.get("initial", {})
            kwargs["initial"]["recipients"] = list(
                set(
                    kwargs["initial"].get("recipients", [])
                    + [r[0] for r in instance.recipients.values_list("recipient_id")]
                )
            )
        super().__init__(*args, **kwargs)

    def _save_m2m(self):
        super()._save_m2m()
        recipients = set(map(int, self.cleaned_data["recipients"]))
        for recipient in self.instance.all_recipients:
            if recipient.recipient_id not in recipients:
                recipient.delete()
        for user_id in recipients:
            MessageRecipient.objects.get_or_create(recipient_id=user_id, message=self.instance)
        return self.instance



# Generated by Django 3.2.25 on 2025-04-04 06:41

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0086_attachment_events'),
    ]

    operations = [
        migrations.AlterField(
            model_name='schoolyear',
            name='year',
            field=models.IntegerField(help_text='Only the first of the two years.', unique=True, validators=[django.core.validators.MinValueValidator(2000), django.core.validators.MaxValueValidator(3000)], verbose_name='year'),
        ),
    ]
